import numpy as numpy
import math

from .ConvergenceError import ConvergenceError
from .BaseFitter import BaseFitter
from .IterationPlotter import IterationPlotter
from .GaussErrorDistribution import GaussErrorDistribution
from .LaplaceErrorDistribution import LaplaceErrorDistribution
from .CauchyErrorDistribution import CauchyErrorDistribution
from .PoissonErrorDistribution import PoissonErrorDistribution
from .ExponentialErrorDistribution import ExponentialErrorDistribution
from .Formatter import formatter as fmt

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2023 Do Kester

class IterativeFitter( BaseFitter ):
    """
    Base class with methods common to all iterative fitters.

    Author:      Do Kester.

    Attributes
    ----------
    tolerance : float
        When absolute and relative steps in subsequent chisq steps are less than
        tolerance, the fitter stops. Default = 0.0001
    maxIter : int
        When the number of iterations gets larger than maxiter the fitter
        stops with a ConvergenceError Default = 1000 * nparams
    iter : int (read only)
        iteration counter
    ntrans : int (read only)
        number of transforms
    verbose : int
        information per iteration.
        0 : silent
        1 : base information (default)
        2 : report about every 100th iteration
        3 : report about every ietration
    tooLarge : int
        When the length parameter array is too large to make a Hessian.
        To avert OutOfMemory. Default = 100

    plotter : Plotter
        Iteration plotter class. Default = IterationPlotter
    plotIter : int
        Produce a plot for every plotIter-th iteration.
        Default = 0 (no plotting)

    Raises
    ------
    ConvergenceError    Something went wrong during the convergence if the fit.

    """
    def __init__( self, xdata, model, maxIter=None, tolerance=0.0001, verbose=1, **kwargs ) :

        """
        Create a new iterative fitter, providing xdatas and model.

        This is a base class. It collects stuff common to all iterative fitters.
        It does not work by itself.

        Parameters
        ----------
        xdata : array_like
            array of independent input values
        model : Model
            the model function to be fitted

        tolerance : float
            When absolute and relative steps in subsequent chisq steps are less than
            tolerance, the fitter stops. Default = 0.01
        maxIter : None or int
            When the number of iterations gets larger than maxiter the fitter
            stops with a ConvergenceError Default = 1000 * nparams
        verbose : int
            0 : silent
            1 : report result
            2 : report every 100th iteration
            3 : report every iteration
        kwargs for @BaseFitter
            map, keep, fixedScale

        """
        super( IterativeFitter, self ).__init__( xdata, model, **kwargs )

        self.tolerance = tolerance
        if maxIter is None :
            maxIter = 1000 * model.npchain
        self.maxIter = maxIter
        self.verbose = verbose
        self.tooLarge = 100
        self.plotter = IterationPlotter()

        self.plotfreq = 0

    #  *************************************************************************
    def setParameters( self, params ):
        """
        Initialize the parameters of the model
        A little superfluous: see {@link Model#setParameters}

        Parameters
        ----------
        params : array_like
            initial parameters

        """
        self.model.parameters = params

    def doPlot( self, param, force=False ):
        """
        Plot intermediate result.

        Parameters
        ----------
        param : array_like
            of the model
        force : bool
            do the plot

        """
        if ( self.plotfreq > 0 and self.iter % self.plotfreq == 0 ) or force :
            y = self.model.result( self.xdata, param )
            self.plotter.plotResult( self.xdata, y, self._iter )

    def fitprolog( self, ydata, weights=None, accuracy=None, keep=None ) :
        """
        Prolog for all iterative Fitters.

        1. Sets up plotting (if requested)
        2. Sets self.iter and self.ntrans to 0
        3. Checks data/weighs for Nans
        4. Makes fitIndex.

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted
        weights : array_like
            weights pertaining to the data
        accuracy : float or array_like
            accuracy of (individual) data
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)

        Returns
        -------
        fitIndex : ndarray of int
            Indices of the parameters that need fitting

        """
        if self.plotfreq > 0 :
            self.plotter.plotData( self.xdata, ydata, self.__str__( ) )

        self.iter = 0
        self.ntrans = 0

        return super( ).fitprolog( ydata, weights=weights, accuracy=accuracy, keep=keep )

    #  *************************************************************************
    def fit( self, ydata, weights=None, keep=None, **kwargs ):
        """
        Return model parameters fitted to the data.

        It will calculate the hessian matrix and chisq.

        Parameters
        ----------
        ydata : array_like
            the data vector to be fitted
        weights : array_like
            weights pertaining to the data
        keep : dict of {int:float}
            dictionary of indices (int) to be kept at a fixed value (float)
        kwargs :
            passed to the fitter

        Raises
        ------
        ConvergenceError if it stops when the tolerance has not yet been reached.

        """
        raise ConvergenceError( "IterativeFitter is a base class, not suitable itself to perform fits." )

    def report( self, verbose, param, chi, more=None, force=False ) :
        """
        Report on intermediate results.
        """

        if verbose > 1 and ( self.iter % 100 == 0 or force ) :
            mr = "" if more is None else fmt( more, format='   %6.1f ' )
            mx = 5 if verbose < 4 else None if verbose == 4 else verbose
            print( fmt( self.iter, format='%6d' ), mr, fmt( chi, format="%8.1f " ),
                   fmt( param, max=mx ) )



    def __str__( self ):
        return "IterativeFitter"





# run with : python3 -m unittest TestFitter

import numpy as numpy
import os
from numpy.testing import assert_array_almost_equal as assertAAE
from astropy import units
import unittest
import FitPlot

from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2004 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class TestQRFitter( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Compare results of QRFitter with Fitter.

    Author:      Do Kester

    """
    aa = 5              #  average offset
    bb = 2              #  slope
    ss = 0.3            #  noise Normal distr.

    x = numpy.asarray( [ -1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
    x += 2
    noise = numpy.asarray( [-0.000996, -0.046035,  0.013656,  0.418449,  0.0295155,  0.273705,
      -0.204794,  0.275843, -0.415945, -0.373516, -0.158084] )

    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def eq( self, a, b, eps=1e-10 ) :
        if ( a + b ) != 0 :
            return abs( a - b ) / abs( a + b ) < eps
        else :
            return abs( a - b ) < eps



    #  **************************************************************
    def testStraightLineParameter( self ):
        """
        test slope fit

        1. Compare PolynomialModel(1) with chain of PowerModel(0) + PowerModel(1)

        2. Fix parameters in PolynomialModel

        """
        print( "\n   Fitter Test 1  \n" )
        plot = self.doplot

        model = PolynomialModel( 1 )
        fitter = Fitter( self.x, model )
        model0 = PowerModel( 0 )
        model1 = PowerModel( 1 )
        model0.addModel( model1 )
        altfit = QRFitter( self.x, model0 )
        yy = self.aa + self.bb * self.x
        y = yy + self.noise

        print( "Testing Straight Line fit" )

        self.assertTrue( model1.npchain == 2 )
        par = fitter.fit( y )
        alt = altfit.fit( y )

        print( "offst = %f  alt = %f  truth = %f"%(par[0], alt[0], self.aa) )
        print( "slope = %f  alt = %f  truth = %f"%(par[1], alt[1], self.bb) )
        assertAAE( par, alt )

        chisq = fitter.chisq
        altch = altfit.chisq
        print( "chisq = %f  alt = %f"%( chisq, altch) )
        self.assertTrue( self.eq(chisq, altch) )

        std = fitter.standardDeviations
        ast = altfit.standardDeviations
        print( "stdev = %f  alt = %f"%(std[0], ast[0]) )
        print( "stdev = %f  alt = %f"%(std[1], ast[1]) )
        assertAAE( std, ast )

        print( "   par  ", par, " stdv  ", std, chisq )

        numpy.random.seed( 9876 )
        for k in range( 5 ) :
            y1 = yy + 0.3 * numpy.random.randn( 11 )
            par1 = fitter.fit( y1 )
            std1 = fitter.stdevs
            chi1 = fitter.chisq
            print( k, " par  ", par1, " stdv  ", std1, chi1 )

        altfit.needsNewDecomposition = True
        par1 = altfit.fit( y, keep={0:par[0]} )
        print( par1, par )
        std1 = altfit.stdevs
        print( std, std1 )
        if plot :
            FitPlot.plotFit( self.x, y, model0 )

        self.assertTrue( 2 == len( par1) )
        assertAAE( par, par1 )

        self.assertTrue( model0.npchain == 2 )
        mp = model0.parameters
        print( mp, par, par1 )
        assertAAE( mp, par1 )

        altfit = QRFitter( self.x, model0, keep={0:par[0]+0.001} )


        par1 = altfit.fit( y )
        std1 = altfit.stdevs
        print( par, par1 )
        print( std, std1 )
        if plot :
            FitPlot.plotFit( self.x, y, model0 )

        self.assertTrue( 2 == len( par1) )
        assertAAE( par, par1, 3 )
        self.assertTrue( std1[0] == 0 )

        error1 = altfit.monteCarloError( )
        print( "error = ", error1 )


if __name__ == '__main__':
    unittest.main( )






import numpy as numpy
import math
import warnings

from .ErrorDistribution import ErrorDistribution
from .LogFactorial import logFactorial

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2023 Do Kester


class PoissonErrorDistribution( ErrorDistribution ):
    """
    To calculate a Poisson likelihood.

    For one observation with n counts it holds

        f( n,x ) = x^n / ( e^x * n! )

    where x is the expected counts

    The function is mostly used to calculate the likelihood L, or easier
    to use log likelihood, logL.

        logL = &sum;( n * log( x ) - x - log( n! ) )

    Weights are not accepted in this ErrorDistribution; they are silently ignored.


    Author       Do Kester.

    """

    PARNAMES = []

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, copy=None ):
        """
        Constructor.

        Parameters
        ----------
        copy : PoissonErrorDistribution
            distribution to be copied.

        """
        super( PoissonErrorDistribution, self ).__init__( copy=copy )

    def copy( self ):
        """ Return copy of this.  """
        return PoissonErrorDistribution( copy=self )

    def acceptWeight( self ):
        """
        True if the distribution accepts weights.
        Always false for this distribution.
        """
        return False


    def getScale( self, problem, allpars=None ) :
        """
        Return the noise scale.

        *** Gaussian approximation ***

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            None take parameters from problem.model
            list of all parameters in the problem
        """
        return self.getGaussianScale( problem, allpars=allpars )

    #  *********LIKELIHOODS***************************************************
    def logLikelihood_alt( self, problem, allpars ):
        """
        Return the log( likelihood ) for a Poisson distribution.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem

        """
        self.ncalls += 1

        mock = problem.result( allpars )
        if numpy.any(  mock <= 0.0 ) :
            return -math.inf

        lfdata = logFactorial( problem.ydata )

        logl = numpy.sum( problem.ydata * numpy.log( mock ) - mock - lfdata )

        if math.isnan( logl ) :
            return -math.inf

        return logl

    def logLdata( self, problem, allpars, mockdata=None ) :
        """
        Return the log( likelihood ) for each residual

        logL = sum( logLdata )

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem
        mockdata : array_like
            as calculated by the model

        """
        if mockdata is None :
            mockdata = problem.result( allpars )
        lfdata = logFactorial( problem.ydata )

        with warnings.catch_warnings():
            warnings.simplefilter( "ignore", category=RuntimeWarning )
            lld = problem.ydata * numpy.log( mockdata ) - mockdata - lfdata

        lld = numpy.where( numpy.isfinite( lld ), lld, -math.inf )

#        lld = numpy.where( mockdata <= 0, -math.inf,
#                problem.ydata * numpy.log( mockdata ) - mockdata - lfdata )


        return lld

    def partialLogL_alt( self, problem, allpars, fitIndex ):
        """
        Return the partial derivative of log( likelihood ) to the parameters.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem
        fitIndex : array_like
            indices of parameters to be fitted
        """
        self.nparts += 1
        mock = problem.result( allpars )
        dM = problem.partial( allpars )
        dL = numpy.zeros( len( fitIndex ), dtype=float )

        i = 0
        for k in fitIndex :
            dL[i] = numpy.sum( ( problem.ydata / mock - 1 ) * dM[:,k] )
            i += 1

        return dL

    def nextPartialData( self, problem, allpars, fitIndex, mockdata=None ):
        """
        Return the partial derivative of log( likelihood ) to the parameters.

        Parameters
        ----------
        problem : Problem
            to be solved
        allpars : array_like
            list of all parameters in the problem
        fitIndex : array_like
            indices of parameters to be fitted
        mockdata : array_like
            as calculated by the model
        """
        if mockdata is None :
            mockdata = problem.result( allpars )
        dM = problem.partial( allpars )
##      TBD import mockdata into partial
#        dM = problem.partial( allpars, mockdata=mockdata )

        for k in fitIndex :
            yield ( problem.ydata / mockdata - 1 ) * dM[:,k]

        return

    def __str__( self ) :
        return "Poisson error distribution"




# run with : python3 -m unittest TestFitter

import numpy as numpy
from numpy.testing import assert_array_almost_equal as assertAAE
from astropy import units
import unittest
import os

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2004 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2017 Do Kester

class TestFitter( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author:      Do Kester

    """
    aa = 5              #  average offset
    bb = 2              #  slope
    ss = 0.3            #  noise Normal distr.

    x = numpy.asarray( [ -1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
    x += 2
    noise = numpy.asarray( [-0.000996, -0.046035,  0.013656,  0.418449,  0.0295155,  0.273705,
      -0.204794,  0.275843, -0.415945, -0.373516, -0.158084] )

    def eq( self, a, b, eps=1e-10 ) :
        if ( a + b ) != 0 :
            return abs( a - b ) / abs( a + b ) < eps
        else :
            return abs( a - b ) < eps

    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    #  **************************************************************
    def testStraightLineParameter( self ):
        """
        test slope fit

        1. Compare PolynomialModel(1) with chain of PowerModel(0) + PowerModel(1)

        2. Fix parameters in PolynomialModel

        """
        plot = self.doplot

        print( "\n   Fitter Test 1  \n" )
        model = PolynomialModel( 1 )
        fitter = Fitter( self.x, model )
        model0 = PowerModel( 0 )
        model1 = PowerModel( 1 )
        model0.addModel( model1 )
        altfit = Fitter( self.x, model0 )
        y = self.noise + self.aa + self.bb * self.x

        print( "Testing Straight Line fit" )

        self.assertTrue( model1.npchain == 2 )
        par = fitter.fit( y )
        alt = altfit.fit( y )

        plotFit( self.x, data=y, model=model, fitter=fitter, residuals=True, show=self.doplot )


        print( "offst = %f  alt = %f  truth = %f"%(par[0], alt[0], self.aa) )
        print( "slope = %f  alt = %f  truth = %f"%(par[1], alt[1], self.bb) )
        assertAAE( par, alt )

        chisq = fitter.chisq
        altch = altfit.chisq
        print( "chisq = %f  alt = %f"%( chisq, altch) )
        self.assertTrue( self.eq(chisq, altch) )

        std = fitter.standardDeviations
        ast = altfit.standardDeviations
        print( "stdev = %f  alt = %f"%(std[0], ast[0]) )
        print( "stdev = %f  alt = %f"%(std[1], ast[1]) )
        assertAAE( std, ast )

        par1 = altfit.fit( y, keep={0:par[0]} )

        plotFit( self.x, data=y, model=model0, fitter=altfit, residuals=False, 
            figsize=[12,7], xlim=[0,12], ylim=[-10,10], show=self.doplot )

        self.assertTrue( 2 == len( par1) )
        assertAAE( par1, par )
        self.assertTrue( model0.npchain == 2 )

        model1 = PolynomialModel( 2, fixed={2:0.0} )

        altfit = Fitter( self.x, model1 )

        par1 = altfit.fit( y )
        print( par, par1 )

        print( "x  ", self.x )
        print( "y  ", y )

        plotFit( self.x, data=y, model=model0, show=self.doplot )

        self.assertTrue( 2 == len( par1) )
        assertAAE( par, par )

        error1 = altfit.monteCarloError( )
        print( "error = ", error1 )

    #  **************************************************************
    def testNormalize( self ):
        """
        test normalized parameter, alternative for keepfixed.

        1.

        2.

        """
        plot = self.doplot

        print( "\n   Fitter Test 2 Normalize  \n" )
        model = PolynomialModel( 2 )

        sm = PowerModel( 1 )

        numpy.random.seed( 2345 )
        x = numpy.linspace( 0.0, 10.0, 101 )
        y = 1.0 + 0.5 * x - 0.2 * x * x +  numpy.random.randn( 101 ) * 0.1

        sm += model                     ## degenerate model par[0] == par[2]

        print( sm )

        fitter = Fitter( x, sm )

        self.assertTrue( sm.npchain == 4 )

        print( fmt( fitter.hessian ) )

        ppp = fitter.fit( y, plot=self.doplot )
        print( ppp )
#        self.assertRaises( numpy.linalg.linalg.LinAlgError, fitter.fit, y )

        fitter.normalize( [0.0, 0.0, 1.0, 0.0], 1.0 )       ## fix par[2] at 1.0

        print( fmt( fitter.hessian ) )
        print( fmt( fitter.covariance ) )

        par = fitter.fit( y, plot=self.doplot )

        print( "param = ", sm.parameters )
        assertAAE( par[2], 1.0 )
        assertAAE( par, [ -0.5, 1.0, 1.0, -0.2], 1 )

        chisq = fitter.chisq
        print( "chisq = %f"%( chisq) )

        std = fitter.standardDeviations
        print( "stdev = ", std )

#        assertAAE( std, ast )


    def test3( self ):

        p = [3.2, -0.1, 0.3, 1.1, 2.1]
        ss = 0.3

        ndata = 101
        x = numpy.linspace( -1, +1, ndata, dtype=float )
        y = ( x - p[1] ) / p[2]
        numpy.random.seed( 3456 )
        y = p[0] * numpy.exp( -y * y ) + ss * numpy.random.randn( ndata )
        y += ( p[3] + p[4] * x )
        p1 = p



        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )
        print( "Testing Linear Fitter with limits" )
        print( "++++++++++++++++++++++++++++++++++++++++++++++++++" )


        print( "   1. no limits and keep in fit " )
        modl1 = PolynomialModel( 4 )

        amfit = Fitter( x, modl1 )

        par0 = amfit.limitsFit( amfit.fit, y )
        print( "pars0   ", fmt( par0 ) )
        print( "stdv0   ", fmt( amfit.stdevs ) )
        print( "chisq0  ", amfit.chisq )

        par1 = amfit.limitsFit( amfit.fit, y, keep={0:3.3} )
        print( "pars1   ", fmt( par1 ) )
        print( "stdv1   ", fmt( amfit.stdevs ) )
        print( "chisq1  ", amfit.chisq )

        printclass( amfit )

        print( "   2. limits and keep in fit " )
        modl2 = PolynomialModel( 4 )
        modl2.setLimits( [-7.0], [7.0] )

        lmfit = Fitter( x, modl2 )

        par2 = lmfit.limitsFit( lmfit.fit, y )
        print( "pars2   ", fmt( par2 ) )
        print( "stdv2   ", fmt( lmfit.stdevs ) )
        print( "chisq2  ", lmfit.chisq )

        par3 = lmfit.limitsFit( lmfit.fit, y, keep={0:3.3} )
        print( "pars3   ", fmt( par3 ) )
        print( "stdv3   ", fmt( lmfit.stdevs ) )
        print( "chisq3  ", lmfit.chisq )

        printclass( lmfit )

        print( "   3. limits and keep in fitter and fit " )
        modl3 = PolynomialModel( 4 )
        modl3.setLimits( [-7.0], [7.0] )

        lmfit = Fitter( x, modl3, keep={4:5.0} )

        par4 = lmfit.limitsFit( lmfit.fit, y )
        print( "pars4   ", fmt( par4 ) )
        print( "stdv4   ", fmt( lmfit.stdevs ) )
        print( "chisq4  ", lmfit.chisq )

        par5 = lmfit.limitsFit( lmfit.fit, y, keep={0:3.3} )
        print( "pars5   ", fmt( par5 ) )
        print( "stdv5   ", fmt( lmfit.stdevs ) )
        print( "chisq5  ", lmfit.chisq )

        printclass( lmfit )

        if self.doplot :
            xx = numpy.linspace( -1, +1, 1001 )
            plt.plot( x, y, 'k+' )
            plt.plot( xx, modl1.result( xx ), 'k-' )

            plt.plot( xx, modl2.result( xx ), 'r-' )
            plt.show()



if __name__ == '__main__':
    unittest.main( )






# run with : python3 -m unittest TestPhantomSampler
# or :       python3 -m unittest TestPhantomSampler.TestPhantomSampler.test1

import unittest
import os
import time
import numpy as numpy
from astropy import units
import math
from numpy.testing import assert_array_almost_equal as assertAAE
from FitPlot import plotFit

from BayesicFitting import *
from BayesicFitting import formatter as fmt
from BayesicFitting import fma
#from BayesicFitting import PhantomSampler1 as PhantomSampler

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestPhantomSampler( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """

    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def makeData( self, n=3 ) :

        N = 201
        a0 = 8.0
        x0 = 0.7
        s0 = 0.4
        b0 = 1.0
        b1 = 0.2
        p0 = 1.3
        cs = 0.3
        cc = 0.2
        r0 = 0.1

        x = numpy.arange( N, dtype=float ) / 25 - 2
        w = numpy.full( N, 2.0, dtype=float )

        g = ( x - x0 ) / s0
        g *= -0.5 * g
        g = a0 * numpy.exp( g )

        b = b1 * x + b0

        s = 2 * math.pi * p0 * x
        s = cs * numpy.sin( s ) + cc * numpy.cos( s )

        numpy.random.seed( 13456 )

        r = r0 * numpy.random.randn( N )
        if n == 1 :
            y0 = g
            pp = numpy.asarray( [a0,x0,s0], dtype=float )
        elif n == 2 :
            y0 = g + b
            pp = numpy.asarray( [a0,x0,s0,b0,b1], dtype=float )
        else :
            y0 = g + b + s
            pp = numpy.asarray( [a0,x0,s0,b0,b1,p0,cc,cs], dtype=float )

        y = y0 + r

        return pp, y0, x, y, w


    def test1( self ):
        print( "=========== Phantom Sampler test 1 ======================" )

        plot = self.doplot

        pp, y0, x, y, w = self.makeData( n=1 )

        gm = GaussModel( )

        print( gm.shortName( ) )

        lolim = numpy.asarray( [-10,-10,  0], dtype=float )
        hilim = numpy.asarray( [ 10, 10, 10], dtype=float )

        gm.setLimits( lolim, hilim )

        ns = PhantomSampler( x, gm, y, w, threads=True, engines=["galilean", "gibbs"] )
#        ns.verbose = 4

        self.dofit( ns, pp, plot=plot )

        sl = ns.samples

        printclass( sl[0] )

#        start = time.time()
#        SampleMovie( sl, problem=ns.problem, kpar=[1,0] )
#        endt = time.time()

#        print( "Elapsed ", endt - start )

    def test2a( self ):
        print( "=========== Phantom Sampler test 2a ======================" )

        plot = self.doplot

        pp, y0, x, y, w = self.makeData( 2 )

        x = numpy.append( x, x )
        y = numpy.append( y, y )

        gm = GaussModel( )
        gm.addModel( PolynomialModel(1) )

        print( gm.shortName( ) )
        print( gm._next.shortName( ) )

        lolim = numpy.asarray( [-10,-10,  0,-10,-10], dtype=float )
        hilim = numpy.asarray( [ 10, 10, 10, 10, 10], dtype=float )

        gm.setLimits( lolim, hilim )

        lmf = LevenbergMarquardtFitter( x, gm )
        pars = lmf.fit( y )
        print( "LMFpars ", fmt( pars, max=None ) )
        print( "LMFstdv ", fmt( lmf.stdevs, max=None ) )

        ns = PhantomSampler( x, gm, y )
        ns.verbose = 2

        evi = ns.sample()
        print( "NS pars ", fmt( ns.parameters ) )
        print( "NS stdv ", fmt( ns.stdevs ) )
        print( "NS scal ", fmt( ns.scale ) )

    def test2b( self ):
        print( "=========== Phantom Sampler test 2b ======================" )

        plot = self.doplot

        pp, y0, x, y, w = self.makeData( 2 )

        x = numpy.append( x, x )
        y = numpy.append( y, y )

        gm = GaussModel( )
        gm.addModel( PolynomialModel(1) )

        print( gm.shortName( ) )
        print( gm._next.shortName( ) )

        lolim = numpy.asarray( [-10,-10,  0,-10,-10], dtype=float )
        hilim = numpy.asarray( [ 10, 10, 10, 10, 10], dtype=float )

        gm.setLimits( lolim, hilim )
        ns = PhantomSampler( x, gm, y )
        ns.verbose = 2
        ns.distribution.setLimits( [0.01, 100] )

        evi = ns.sample()
        print( "NS pars ", fmt( ns.parameters ) )
        print( "NS stdv ", fmt( ns.stdevs ) )
        print( "NS scal ", fmt( ns.scale ) )


    def test4( self ):
        print( "=========== Phantom Sampler test 4 ======================" )

        pp = 0.0
        numpy.random.seed( 13456 )

        for N in [10, 40, 160, 640, 2560] :
            print( "======= ", N, " points ======" )
            x = numpy.linspace( -1.0, 1.0, N, dtype=float )
            y = numpy.random.randn( N )

            pm = PolynomialModel( 1 )

            lolim = numpy.asarray( [-10], dtype=float )
            hilim = numpy.asarray( [ 10], dtype=float )

            pm.setLimits( lolim, hilim )

            lmf = LevenbergMarquardtFitter( x, pm )
            pars = lmf.fit( y )
            lmfevid = lmf.getEvidence( limits=[lolim,hilim], noiseLimits=[0.01,100] )
            print( "LMFpars ", fmt( pars, max=None ) )
            print( "LMFstdv ", fmt( lmf.stdevs, max=None ) )
            print( "LMFscal ", fmt( lmf.scale ) )
            print( "LMFevid ", fmt( lmfevid ) )

            ns = PhantomSampler( x, pm, y, verbose=0 )
            ns.distribution.setLimits( [0.01, 100] )

            evi = ns.sample()
            print( "NS pars ", fmt( ns.parameters ) )
            print( "NS stdv ", fmt( ns.stdevs ) )
            print( "NS scal ", fmt( ns.scale ) )
            print( "NS evid ", fmt( evi ), " +- ", fmt( ns.precision ) )


    ### Does not always fit properly 
    def XXXtest3( self ):
        print( "=========== Phantom Sampler test 3 ======================" )

        plot = self.doplot

        pp, y0, x, y, w = self.makeData( 3 )

        gm = GaussModel( )
        gm.addModel( PolynomialModel(1) )
        gm.addModel( SineModel() )

        print( gm.shortName( ) )
        print( gm._next.shortName( ) )
        print( gm._next._next.shortName( ) )
        print( gm._next._next._next )

        lolim = numpy.asarray( [-10,-10,  0,-10,-10, 0.5,-10,-10], dtype=float )
        hilim = numpy.asarray( [ 10, 10, 10, 10, 10, 2.5, 10, 10], dtype=float )

        gm.setLimits( lolim, hilim )

        ns = PhantomSampler( x, gm, y, w, seed=89071 )
        ns.distribution.setLimits( [0.01, 100] )

        Tools.printclass( ns )

        print( "truth  ", pp )
        self.dofit( ns, pp, plot=plot )


    def dofit( self, ns, pp, plot=False ) :
        logE = ns.sample( plot=plot )

        par = ns.parameters
        std = ns.standardDeviations
        mlp = ns.samples.maxLikelihoodParameters
        scale = ns.scale
        scdev = ns.stdevScale
        print( "truth  ", fma( pp ) )
        print( "par    ", fma( par ) )
        print( "st dev ", fma( std ) )
        print( "ML par ", fma( mlp ) )
        print( "scale  ", fmt( scale ) )
#        ns.report( )


        self.assertTrue( all( numpy.abs( par - mlp ) < 2 * std ) )
        self.assertTrue( all( numpy.abs( par - pp ) < 2 * std ) )
#        assertAAE( par, pp, 2 )

    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( PhantomSampler1Test.__class__ )





import numpy as numpy
import math as math
import warnings

from .Tools import setAttribute as setatt
from .Tools import printclass

__author__ = "Do Kester"
__year__ = 2022
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"


#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2022 Do Kester

class Prior( object ):
    """
    Base class defining prior distributions.

    Two methods need to be defined in specific priors which map
    the values between [0,1] on to the domain, and vice versa:
    unit2Domain and domain2Unit.

        u = domain2Unit( d )
        d = unit2Domain( u )

    d is a value in the domain of the prior and u is a vlue in [0,1]

    The handling of limits is relegated to this Prior class. Define
        _umin = domain2Unit( lowLimit )
        _urng = domain2Unit( highLimit ) - umin

        u = ( domain2Unit( d ) - umin ) / urange
        d = unit2Domain( u * urange + umin )

    Symmetric priors can be used in a circular variant; i.e.
    the low and high limits are folded on each other, provided
    that the limit values are the same (hence symmetric)

        u = limitedDomain2Unit( d ) + 1 ) / 3
        d = limitedUnit2Domain( ( u * 3 ) % 1 )

    The copy method is also necessary.

    Attributes
    ----------
    lowLimit : float
        low limit on the Prior
    highLimit : float
        high limit on the Prior
    deltaP : float
        width of numerical partial derivative calculation
    circular : bool or float
        whether circular

    Hidden Attributes
    -----------------
    _lowDomain : float
        lower limit of the Priors possible values
    _highDomain : float
        upper limit of the Priors possible values
    _umin : float
        umin lowLimit in unit
    _urng : float
        urange (hi-lo) in unit
    """

##  TBD : 1. change all self.attribute = value to setattr( "attribute, value )
##        2. forbid external settings of (most) attributes except through methods


    #*********CONSTRUCTORS***************************************************
    def __init__( self, limits=None, circular=False, domain=[-math.inf,math.inf],
                  prior=None ):
        """
        Default constructor.

        Parameters
        ----------
        limits : None or list of 2 floats
            2 limits resp. low and high
        circular : bool or float
            False not circular
            True  circular with period from limits[0] to limits[1]
            period of circularity
        domain : 2 floats
            over which the prior is defined
        prior : Prior
            prior to copy (with new limits if applicable)
        """
        super( object, self ).__init__()

        self.deltaP = 0.0001                # for numerical partials
        # private properties of the Prior
        self._lowDomain  = domain[0]         # Lower limit of the Priors possible domain.
        self._highDomain = domain[1]         # Upper limit of the Priors possible domain

        self.setPriorAttributes( limits, circular )

        if prior is not None :
            self.deltaP = prior.deltaP

    def copy( self ) :
        """ Return a copy """
        return Prior( prior=self, limits=self.limits, circular=self.circular )

    def setLimits( self, limits=None ):
        """
        Set limits.
        It is asserted that lowLimit is smaller than highLimit.

        Parameters
        ----------
        limits : None or list of any combination of [None, float]
            None : no limit (for both or one)
            float : [low,high] limit

        Raises
        ------
        ValueError when low limit is larger than high limit or out of Domain

        """
#        print( "Prior 1 ", limits )
        if limits is None :
            self.lowLimit  = self._lowDomain
            self.highLimit = self._highDomain
            reset = False
        else :
            self.lowLimit  = limits[0] if limits[0] is not None else self._lowDomain
            self.highLimit = limits[1] if limits[1] is not None else self._highDomain
            reset = ( limits[0] is not None ) or ( limits[1] is not None )

#        print( "Prior 2 ", self.lowLimit, self.highLimit )
        if not ( self._lowDomain <= self.lowLimit < self.highLimit <= self._highDomain ) :
            print( self._lowDomain, self.lowLimit, self.highLimit, self._highDomain )
            raise ValueError( "Limits out of order or out of domain" )

        try :
            self._umin = 0.0            ## only for re-setting the limits
            self._urng = 1.0
            umax = self.domain2Unit( self.highLimit )
            self._umin = self.domain2Unit( self.lowLimit )
            self._urng = umax - self._umin
        except :
            self._urng = math.inf

#        print( "Prior 3 ", self._umin, self._urng )

        if math.isinf( self._urng ) :
#            warnings.warn( "%s needs limits" % self.__str__() )
            return

        if hasattr( self, "baseUnit2Domain" ) : return			# aleady reset

        # reset the functions
        if reset :
            setatt( self, "baseUnit2Domain", self.unit2Domain )
            setatt( self, "unit2Domain", self.limitedUnit2Domain )
            setatt( self, "baseDomain2Unit", self.domain2Unit )
            setatt( self, "domain2Unit", self.limitedDomain2Unit )

    def setPriorAttributes( self, limits, circular ) :

        if isinstance( circular, bool ) :
            self.setLimits( limits )
            if circular and limits is not None :
                period = limits[1] - limits[0]
            else :
                return
        else :
            period = circular

            if hasattr( self, "center" ) :
                self.setLimits( limits=[self.center - period / 2, self.center + period / 2] )
            else :
                self.setLimits( limits=[0,period] )

        self.circular = circular

        # reset the functions to circular variant
        setatt( self, "unit2Domain", self.circularUnit2Domain )
        setatt( self, "domain2Unit", self.circularDomain2Unit )


    def isCircular( self ) :
        """
        Whether circular
        """
        return hasattr( self, "circular" ) and self.circular

    def limitedDomain2Unit( self, dval ) :
        """
        Shrink domain to value in [0,1]
        """
        return ( self.baseDomain2Unit( dval ) - self._umin ) / self._urng

    def limitedUnit2Domain( self, uval ) :
        """
        Expand value in [0,1] to domain
        """
        return self.baseUnit2Domain( uval * self._urng + self._umin )

    def circularDomain2Unit( self, dval ) :
        return ( self.limitedDomain2Unit( dval ) + 1 ) / 3

    def circularUnit2Domain( self, uval ) :
        return self.limitedUnit2Domain( ( uval * 3 ) % 1 )

    def __setattr__( self, name, value ) :
        """
        Set attributes: lowLimit, highLimit, deltaP, _lowDomain, _highDomain.

        Also set scale for Priors that need a scale.

        """
        keys = ["lowLimit", "highLimit", "deltaP", "center", "scale",
                "_lowDomain", "_highDomain", "_umin", "_urng", "circular"]

        if name == "scale"  and value <= 0 :
            raise ValueError( "Attribute %s must be positive" % name )

        if name in keys :
            setatt( self, name, value )
        else :
            raise AttributeError( repr( self ) + " object has no attribute " + name )

    def __getattr__( self, name ) :
        """
        Return (virtual) attribute with name

        Parameters
        ----------
        name : str
            name of attribute
        """
        if name == "lowLimit" :
            return self._lowDomain
        if name == "highLimit" :
            return self._highDomain
        if name == "limits" :
            return self.getLimits()
        if name == "circular" :			## when not defined
            return False
        else :
            raise AttributeError( "Prior: Unknown attribute " + name )

    def unsetLimits( self ):
        """ Remove all limits.  """
        self.lowLimit = self._lowDomain
        self.highLimit = self._highDomain

    def setAttributes( self, limits=None, scale=None ) :
        """
        Set possible attributes for a Prior.

        Parameters
        ----------
        limits : float or None
            [low,high] limit
        scale : float or None
            scale factor
        """
        if limits is not None :
            self.setLimits( limits=limits )
        if scale is not None :
            self.scale = scale

    def isOutOfLimits( self, par ):
        """
        True if the parameter is out of limits

        Parameters
        ----------
        par : float
            the parameter to check

        """
        return ( par < self.lowLimit ) or ( par > self.highLimit )

    def checkLimit( self, par ):
        """
        Check whether the parameter is within limits.

        Parameters
        ----------
        par : float
            the parameter to check

        Raises
        ------
            ValueError when outside limits.

        """
        if self.isOutOfLimits( par ):
            raise ValueError( "Parameter outside supplied limits: %8.2f < %8.2f < %8.2f"%
                            (self.lowLimit, par, self.highLimit) )

    def stayInLimits( self, par ):
        """
        Return lower limit or upper limit when parameter is outside.

        Parameters
        ----------
        par : float
            the parameter to check

        """
        if par < self.lowLimit:
            return self.lowLimit
        if par > self.highLimit:
            return self.highLimit
        return par

    def hasLowLimit( self ):
        """ Return true if the prior has its low limits set.  """
        return self.lowLimit > self._lowDomain

    def hasHighLimit( self ):
        """ Return true if the prior has its high limits set.  """
        return self.highLimit < self._highDomain

    def hasLimits( self ):
        """ Return true if it has any limits.  """
        return self.hasLowLimit() and self.hasHighLimit()

    def getLimits( self ):
        """ Return the limits.  """
        return numpy.array( [ self.lowLimit, self.highLimit ] )

    def getIntegral( self ) :
        """
        Return the integral of the prior over the valid range.

        Default: 1.0 (for bound priors)
        """
        return 1.0

    def getRange( self ):
        """ Return the range.  """
        return self.highLimit - self.lowLimit


    def domain2Unit( self, dval ):
        """
        Return a value in [0,1] given a value within the valid domain of
        a parameter for a distribution.

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        pass

    def unit2Domain( self, uval ):
        """
        Return a value within the valid domain of the parameter given a value
        between [0,1] for a distribution.

        Parameters
        ----------
        uval : float
            value within [0,1]

        """
        pass

    def result( self, p ):
        """
        Return value of the Prior at a given value.

        If result is not defined, fall back to numerical derivative od Domain2Unit.

        Parameters
        ----------
        p : float
            the value

        """
        return self.numPartialDomain2Unit( p )

    def partialDomain2Unit( self, p ):
        """
        Return the derivative of Domain2Unit, aka the result of the distribution at p

        Parameters
        ----------
        p : float
            the value

        """
        return self.result( p )

    def logResult( self, p ) :
        """
        Return the log of the result; -inf when p == 0.

        Parameters
        ----------
        p : float
            the value

        """
        try :
            return numpy.log( self.result( p ) )
        except :
            return -math.inf

    def numPartialDomain2Unit( self, dval ):
        """
        Return a the numeric derivate of the domain2Unit function to dval.

        Parameters
        ----------
        dval : float
            value within the domain of a parameter

        """
        return ( ( self.domain2Unit( dval + self.deltaP ) -
                   self.domain2Unit( dval - self.deltaP ) ) /
                        ( 2 * self.deltaP ) )

    def partialLog( self, p ):
        """
        Return partial derivative of log( Prior ) wrt parameter.
        default numPartialLog

        Parameters
        ----------
        p : float
            the value

        """
        return self.numPartialLog( p )

    def numPartialLog( self, p ):
        """
        Return the numeric partial derivative of log( Prior ) wrt parameter.
        Parameters
        ----------
        p : float
            the value

        """
        if self.isOutOfLimits( p ) : return math.nan

        rm = self.logResult( p - self.deltaP )
        rp = self.logResult( p + self.deltaP )
        return ( rp - rm ) / ( 2 * self.deltaP )

    def isBound( self ):
        """ Return true if the integral over the prior is bound.  """
        return False

    def __str__( self ):
        """ Return a string representation of the prior.  """

        name = self.shortName()

        if hasattr( self, "center" ) :
            name += str( " at (%.2f %.2f)" % ( self.center, self.scale ) )

        if self.hasLowLimit() or self.hasHighLimit():
            name += str( " circular" if self.isCircular() else " with limits" )
            name += str( " between %.2f and %.2f" % ( self.lowLimit, self.highLimit ) )

        return name

#      * End of Prior





from __future__ import print_function

import numpy as numpy
from astropy import units
import math
from . import Tools
from .Tools import setAttribute as setatt
from .Formatter import formatter as fmt
from .Formatter import fma
from . import Plotter
import sys
import warnings
import matplotlib.pyplot as plt

from .NestedSampler import NestedSampler


from .Explorer import Explorer
from .Model import Model
from .Walker import Walker
from .WalkerList import WalkerList
from .Sample import Sample
from .SampleList import SampleList

from .Problem import Problem
from .ClassicProblem import ClassicProblem
from .ErrorsInXandYProblem import ErrorsInXandYProblem

from .ErrorDistribution import ErrorDistribution
from .ScaledErrorDistribution import ScaledErrorDistribution
from .GaussErrorDistribution import GaussErrorDistribution
from .LaplaceErrorDistribution import LaplaceErrorDistribution
from .PoissonErrorDistribution import PoissonErrorDistribution
from .CauchyErrorDistribution import CauchyErrorDistribution
from .UniformErrorDistribution import UniformErrorDistribution
from .ExponentialErrorDistribution import ExponentialErrorDistribution

from .Engine import Engine
from .StartEngine import StartEngine
from .ChordEngine import ChordEngine
from .GibbsEngine import GibbsEngine
from .GalileanEngine import GalileanEngine
from .StepEngine import StepEngine
## for Dynamic Models import the classes
from .BirthEngine import BirthEngine
from .DeathEngine import DeathEngine
## for Modifiable Models:
from .StructureEngine import StructureEngine

__author__ = "Do Kester"
__year__ = 2023
__license__ = "GPL3"
__version__ = "3.1.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2017 - 2023 Do Kester

class PhantomSampler( NestedSampler ):
    """
    PhantomSampler is a version of NestedSampler that removes step walkers from
    the list. It uses a selection of the intermediate (phantom) positions which
    are created inside the engines as valid walker points.
    These phantoms are added to the walkerlist, until it fills the
    ensemble to its initial value.

    Each iteration a percentage of the walkers is used to update the evidence
    and are subsequently transferred to the list of posterior samples.

    In principle it speeds up the calculations by a factor step, of course it
    pays in exploratory power and precision.


    Attributes
    ----------
    step : int (< 10)
        percentage of the walkers to replace

    Attributes from NestedSampler
    -----------------------------
    xdata, model, ydata, weights, problem, distribution, ensemble, discard, rng, seed,
    rate, maxsize, minimumIterations, end, verbose, walkers, samples, engines,
    initialEngine

    Author       Do Kester.

    """
    ENSEMBLE = 100

    #  *********CONSTRUCTORS***************************************************
    def __init__( self, xdata=None, model=None, ydata=None, weights=None,
                accuracy=None, problem=None, distribution=None, limits=None, 
                keep=None, ensemble=100, seed=80409, rate=1.0, engines=None, 
                maxsize=None, threads=False, verbose=1, step=4 ) :
        """
        Create a new class, providing inputs and model.

        Either (model,xdata,ydata) needs to be provided or a completely filled
        problem.

        Parameters
        ----------
        step : int
            percentage of walkers to use

        Parameters from NestedSampler
        -----------------------------
        xdata : array_like
            array of independent input values
        model : Model
            the model function to be fitted
            the model needs priors for the parameters and (maybe) limits
        ydata : array_like
            array of dependent (to be fitted) data
        weights : array_like (None)
            weights pertaining to ydata
        accuracy : float or array_like
            accuracy scale for the datapoints
            all the same or one for each data point
        problem : None or string or Problem
            Defines the kind of problem to be solved.

            None        same as "classic"
            "classic" 	ClassicProblem
            "errors"	ErrorsInXandYProblem
            "multiple"	MultipleOutputProblem

            Problem     Externally defined Problem. When Problem has been provided,
                        xdata, model, weights and ydata are not used.
        keep : None or dict of {int:float}
            None of the model parameters are kept fixed.
            Dictionary of indices (int) to be kept at a fixed value (float).
            Hyperparameters follow model parameters.
            The values will override those at initialization.
            They are used in this instantiation, unless overwritten at the call to sample()
        distribution : None or String or ErrorDistribution
            Defines the ErrorDistribution to be used
            When the hyperpar(s) are not to be kept fixed, they need `Prior` and maybe limits.

            None            same as "gauss"
            "gauss"         GaussErrorDistribution with (fixed) scale equal to 1.0
            "laplace"       LaplaceErrorDistribution with 1 hyperpar scale
            "poisson"       PoissonErrorDistribution no hyperpar
            "cauchy"        CauchyErrorDstribution with 1 hyperpar scale
            "uniform"       UniformErrorDistribution with 1 hyperpar scale
            "exponential"   ExponentialErrorDistribution with 2 hyperpar (scale, power)
            "bernoulli"     BernoulliErrorDistribution no hyperpar

            ErrorDistribution Externally defined ErrorDistribution
        limits : None or [low,high] or [[low],[high]]
            None    no limits implying fixed hyperparameters of the distribution
            low     low limit on hyperpars
            high    high limit on hyperpars
            When limits are set the hyperpars are not fixed.
        ensemble : int
            number of walkers
        seed : int
            seed of random number generator
        rate : float
            speed of exploration
        engines : None or (list of) string or (list of) Engine
            to randomly move the walkers around, within the likelihood bound.

            None        use a Problem defined selection of engines
            "galilean"  GalileanEngine	move forward and mirror on edges
            "chord"     ChordEngine   	select random point on random line
            "gibbs" 	GibbsEngine 	move one parameter at a time
            "step"  	StepEngine    	move all parameters in arbitrary direction

            For Dynamic models only:
            "birth" 	BirthEngine     increase the parameter list of a walker by one
            "death" 	DeathEngine     decrease the parameter list of a walker by one

            For Modifiable models only:
            "struct"    StructureEngine change the (internal) structure.

            Engine      an externally defined (list of) Engine
        maxsize : None or int
            maximum size of the resulting sample list (None : no limit)
        threads : bool (False)
            Use Threads to distribute the diffusion of discarded samples over the available cores.
        verbose : int (1)
            0   silent
            1   basic information
            2   more about every 100th iteration
            3   more about every iteration
            >4  for debugging

        """

        if step <= 10 :
            self.step = step
        else :
            raise ValueError( "Please keep step <= 10" )

        self.initEnsemble = ensemble
        self.usePhantoms = True
#        self.walkers = []

        ## ensemble=None entails a dynamic emsemble value.
        super().__init__( xdata=xdata, model=model, ydata=ydata, weights=weights,
                accuracy=accuracy, problem=problem, distribution=distribution, limits=limits, 
                keep=keep, ensemble=None, seed=seed, rate=rate, engines=engines, maxsize=maxsize,
                threads=threads, verbose=verbose )


    def __getattr__( self, name ) :
        if name == "worst" :
            return int( ( len( self.walkers ) * self.step ) // 100 )
        else :
            return super().__getattr__( name )

    def initSample( self, ensemble=None, keep=None ) :

        return super().initSample( ensemble=self.initEnsemble, keep=keep )


    def updateWalkers( self, explorer, worst ) :
        """
        Update the walkerlist while appending the new (phantom) walkers to the list

        Parameters
        ----------
        explorer : Explorer
            Explorer object
        worst : int
            number of walkers to update
        """
        self.walkers[:worst] = []

        e0 = self.engines[0]
#        print( "updateW  ", worst, len( self.walkers ), fma( e0.unitRange, linelength=200 ) )

        while len( self.walkers ) <= self.initEnsemble :
            # Explore the copied walker(s)
            wlist = [self.rng.randint( worst, self.ensemble )]
            explorer.explore( wlist, self.lowLhood, self.iteration )

#            print( "explore   ", wlist, self.ensemble, min(e0.unitRange), max( e0.unitRange) )


        while len( self.walkers ) > self.initEnsemble :
            ks = self.initEnsemble - self.step
            k = ks + self.rng.randint( self.ensemble - ks - 1 )
#            print( k, self.ensemble, len( self.walkers ), self.walkers[-1].logL )
            del( self.walkers[k] )





import numpy as numpy
import math
from scipy import special

from .NonLinearModel import NonLinearModel
from . import Tools

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2011 Do Kester, SRON (JAVA code)
#  *    2016 - 2021 Do Kester

class VoigtModel( NonLinearModel ):
    """
    Voigt's Gauss Lorentz convoluted model for line profiles.

    The Voigt function is a convolution of a Gauss and a Lorentz function.
    Physicaly it is the result of thermal and pressure broadening of a spectral
    line.

    The models takes 4 parameters: amplitude, center frequency, half-width of
    the Gaussian, and half-width of the Lorentzian.
    These are initialised to [1, 0, 1, 1].
    Parameters 2 & 3 ( widths ) is always kept positive ( >=0 ).

    The implementation uses the Faddeeva function from scipy.special.wofz.

    Examples
    --------
    >>> voigt = VoigtModel( )
    >>> voigt.setParameters( [5, 4, 1, 2] )
    >>> print( voigt( numpy.arange(  41 , dtype=float ) / 5 ) )      # from [0,8]


    Attributes
    ----------
        none of its own

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames


    """

    def __init__( self, copy=None, **kwargs ):
        """
        Voigt model.

        Number of parameters is 4.

        Parameters
        ----------
        copy : VoigtModel
            to be copied

        """
        param = [1.0, 0.0, 1.0, 1.0]
        names = ["amplitude","center","gausswidth","lorentzwidth"]

        super( VoigtModel, self ).__init__( 4, copy=copy, params=param,
                        names=names, **kwargs )
        if copy is None :
            self.posIndex = [2,3]

    def copy( self ):
        """ Copy method.  """
        return VoigtModel( copy=self )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Note: both width in the parameter array ( items 2 & 3 ) are kept
        strictly positive. I.e. they are changed when upon input they are negative.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        sigma = params[2] * math.sqrt( 2.0 )
        gamma = params[3]
        z0 = 1j * gamma / sigma
        z = ( xdata - params[1] ) / sigma + z0


        wofz0 = numpy.real( special.wofz( z0 ) )
        return params[0] * numpy.real( special.wofz( z ) ) / wofz0

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

            z = ( x - p1 + 1j * p3 ) / ( p2 * sqrt2 )
            z0 = 1j * p3 / ( p2 * sqrt2 )

            vgt = p0 * R( wofzz ) / R( wofz0 )

            dvdp = dvdz * dzdp

            dvdz = p0 * ( R(dwdz) * R(wofz0) - R(dwd0) * R(wofzz) ) / R(wofz0)^2
            dvdp = p0 * ( R(dwdz * dzdp) * R(wofz0) - R(dwd0 * d0dp) * R(wofzz) ) / R(wofz0)^2

            dwdz = 2j / sqrt(pi) - 2 * z  * wofzz
            dwd0 = 2j / sqrt(pi) - 2 * z0 * wofz0

            ## p0 and p1 have no influence in wofz0
            dzdp0 = 0
            dzdp1 = -1 / ( p2 * sqrt2 )
            d0dp2 = - ( 1j * p3 / ( p2^2 * sqrt2 )              = -z0 / p2
            dzdp2 = - ( ( x - p1 + 1j * p3 ) / ( p2^2 * sqrt2 ) = -z  / p2
            dzdp3 = d0dp3 = 1j / ( p2 * sqrt2 )

            dvdp0 = R(wofzz) / R(wofz0)
            ## The other partial follow from calculating dvdp for the parameters 1..3

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        partial = numpy.ndarray( ( Tools.length( xdata ), self.npbase ) )

        sigma = params[2] * math.sqrt( 2.0 )
        gamma = params[3]
        z0 = 1j * gamma / sigma
        z = ( xdata - params[1] ) / sigma + z0

        wofz0 = special.wofz( z0 )
        wofzz = special.wofz( z )
        dwd0 = 2j / math.sqrt( math.pi ) - 2 * z0 * wofz0
        dwdz = 2j / math.sqrt( math.pi ) - 2 * z * wofzz

        rwz = numpy.real( wofzz )
        rw0 = numpy.real( wofz0 )
        a = params[0] / rw0
        a2 = a / rw0

        dzdp1 = -1 / sigma
        d0dp2 = -z0 / params[2]
        dzdp2 = -z / params[2]
        dzdp3 = 1j / sigma

        parts = { 0 : ( lambda: rwz / rw0 ),
                  1 : ( lambda: a * numpy.real( dwdz * dzdp1 ) ),
                  2 : ( lambda: a2 * ( numpy.real( dwdz * dzdp2 ) * rw0 -
                                       numpy.real( dwd0 * d0dp2 ) * rwz ) ),
                  3 : ( lambda: a2 * ( numpy.real( dwdz * dzdp3 ) * rw0 -
                                       numpy.real( dwd0 * dzdp3 ) * rwz ) ) }

        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            partial[:,k] = parts[kp]()

        return partial

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each xdata (=x).

            z = ( x - p1 + 1j * p3 ) / ( p2 * sqrt2 )
            z0 = 1j * p3 / ( p2 * sqrt2 )

            vgt = p0 / wofz0 * re( wofzz )
            dvdx = dvdz * dzdx

            dvdz = p0 / wofz0 * dwdz
            dwdz = 2j / sqrt(pi) - 2 * z * wofzz

            dzdx = 1 / ( p2 * sqrt2 )

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        sigma = params[2] * math.sqrt( 2.0 )
        gamma = params[3]
        z0 = 1j * gamma / sigma
        z = ( xdata - params[1] ) / sigma + z0

        wofz0 = numpy.real( special.wofz( z0 ) )
        wofzz = special.wofz( z )
        dwdz = 2j / math.sqrt( math.pi ) - 2 * z * wofzz

        return numpy.real( params[0] * dwdz / ( wofz0 * sigma ) )

    def baseName( self ):
        """
        Returns a string representation of the model.
        """
        return str( "Voigt:" )


    def baseParameterUnit( self, k ):
        """
        Return the name of a parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        if k == 0:
            return self.yUnit
        return self.xUnit





import numpy as numpy
from astropy import units
import math
from . import Tools
from .Tools import setAttribute as setatt
from .Formatter import formatter as fmt

from .Model import Model
from .Dynamic import Dynamic
from .ExponentialPrior import ExponentialPrior
from .UniformPrior import UniformPrior
from .Prior import Prior

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2018 - 2020 Do Kester

class RepeatingModel( Model, Dynamic ):
    """
    RepeatingModel is a dynamic model, that calls the same model zero or more times,
    each time with the next set of parameters.

    RepeatingModel is a Dynamic model that grows and shrinks according to a growPrior.
    The growPrior defaults to ExponentialPrior, unless a maxComp (max nr of components)
    then it defaults to a UniformPrior with limits minComp .. maxComp.

    When isDynamic is set to False or minComp == maxComp, the RepeatingModel is a normal
    model with a static number of parameters/components.

    Priors and/or limits for the RepeatingModel are stored in the (encapsulated) model.

    It can be arranged that all similar parameters are the same, represented by the
    same parameters. Use keywords same=.

    Attributes
    ----------
    ncomp : int
        number of repetitions
    model : Model
        (encapsulated) model to be repeated
    same : None or int or list of int
        indices of parameters of model that get identical values
    index : list of int
        list of parameter indices not in same.
    isDyna : bool
        Whether this is a Dynamic Model.

    Attributes from Dynamic
    -----------------------
        ncomp, deltaNpar, minComp, maxComp, growPrior

    Attributes from Model
    ---------------------
        parameters, stdevs, npchain
        _next, _head, _operation
        xUnit, yUnit (relegated to model)

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    Example
    -------
    >>> # Define a model containing between 1 and 6 VoigtModels, starting with 3
    >>> # and all with the same widths (for Gauss and Cauchy)
    >>> vgt = VoigtModel()
    >>> mdl = RepeatingModel( 3, vgt, minComp=1. maxComp=6, same=[2,3] )
    >>> print( mdl.npbase )             # 4 + 2 + 2
    >>> 8
    >>> # Define a static RepeatingModel of 5 GaussModels
    >>> gm = GaussModel()
    >>> mdl = RepeatingModel( 5, gm, isDynamic=False )
    >>> print( mdl.npbase )             # 5 * 3
    >>> 15
    >>> # Define a RepeatingModel with and exponential grow prior with scale 10
    >>> mdl = RepeatingModel( 1, gm, growPrior=ExponentialPrior( scale=10 ) )
    >>> print( mdl.npbase )             # 3
    >>> 3

    Author       Do Kester

    """
    #  *****CONSTRUCTOR*********************************************************
    def __init__( self, ncomp, model, minComp=0, maxComp=None, fixed=None,
                  same=None, growPrior=None, dynamic=True, copy=None, **kwargs ):
        """
        Repeating the same model several times.

        Parameters
        ----------
        ncomp : int
            number of repetitions
        model : Model
            model to be repeated
        minComp : int (0)
            minimum number of repetitions
        maxComp : None or int
            maximum number of repetitions
        same : None or int or list of int
            indices of parameters of model that get identical values

        growPrior : None or Prior
            governing the birth and death.
            ExponentialPrior (scale=2) if  maxOrder is None else UniformPrior
        dynamic : bool (True)
            Whether this is a Dynamic Model.
        copy : RepeatingModel
            model to copy

        Raises
        ------
        AttributeError when fixed parameters are requested
        ValueError when order is outside [min..max] range

        """
        if fixed is not None :
            raise AttributeError( "DynamicModel cannot have fixed parameters; " +
                                  " put fixed parameter(s0 in model." )
        if ncomp < minComp or ( maxComp is not None and ncomp > maxComp ) :
            raise ValueError( "ncomp outside range of [min..max] range" )

        np = ncomp * model.npchain - ( ncomp - 1 ) * Tools.length( same )
        self.isDyna = dynamic and ( maxComp is None or minComp < maxComp )

        super( RepeatingModel, self ).__init__( np, copy=copy, **kwargs )

        self.ncomp = ncomp
        self.model = model

        if copy is None :
            self.minComp = minComp
            self.maxComp = maxComp
            self.priors = model.priors          ## point to the same
        else :
            self.minComp = copy.minComp
            self.maxComp = copy.maxComp
            growPrior = copy.growPrior.copy()
            same = copy.same

        self.setSame( same )
        self.deltaNpar = len( self.index )

        if self.isDynamic() :
            if growPrior is None :
                if maxComp is None :
                    self.growPrior = ExponentialPrior( scale=2 )
                else :
                    lim = [minComp, maxComp+1]        # limits on components
                    self.growPrior = UniformPrior( limits=lim )
            else :
                self.growPrior = growPrior

    def copy( self ):
        """ Copy method.  """
        return RepeatingModel( self.ncomp, self.model, isDynamic=self.isDyna, copy=self )

    def changeNComp( self, dn ) :
        setatt( self, "ncomp", self.ncomp + dn )

    def setSame( self, same ) :
        """
        Assign similar parameters the same value.

        Parameters
        ----------
        same : None or int or [int]
            similar parameters indicated as an index in encapsulated model.
        """
        self.same = same
        if same is None :
            self.index = numpy.arange( self.model.npbase, dtype=int )
            return

        index = []
        for k in range( self.model.npbase ) :
            if not k in self.same :
                index += [k]
        self.index = numpy.array( index, dtype=int )

    def grow( self, offset=0, rng=None, **kwargs ):
        """
        Increase the the number of components by 1 (if allowed by maxComp)

        Parameters
        ----------
        offset : int
            index where the dynamic model starts
        rng : RandomState
            random numbr generator

        Return
        ------
        bool :  succes

        """
        if not self.isDynamic() :
            return False

        if self.maxComp is not None and self.ncomp >= self.maxComp:
            return False

        location = self.npbase

        dnp = self.deltaNpar if self.ncomp > 0 else self.model.npbase
        self.alterParameterSize( dnp, offset, location=location )

        head = self._head
        mdlpar = head.parameters
        k1 = location + offset
        value = numpy.zeros( dnp, dtype=float )
        if rng is not None :
            for k in range( dnp ) :
                value[k] = head.unit2Domain( rng.rand(), k + k1 )

        mdlpar = self.alterParameters( mdlpar, location, dnp, offset, value=value )

        setatt( self._head, "parameters", mdlpar )

        self.alterParameterNames( dnp )

        self.ncomp += 1

        return True

    def shrink( self, offset=0, **kwargs ):
        """
        Decrease the the number of componenets by 1 (if allowed by minComp)
        Remove an arbitrary item.

        Parameters
        ----------
        offset : int
            index where the dynamic model starts

        Return
        ------
        bool : succes

        """
        if ( not self.isDynamic() ) or self.ncomp <= self.minComp :
            return False

        location = self.npbase
        dnp = -( self.deltaNpar if self.ncomp > 1 else self.model.npbase )
        self.alterParameterSize( dnp, offset, location=location )

        head = self._head
        mdlpar = head.parameters
        mdlpar = self.alterParameters( mdlpar, location, dnp, offset )
        setatt( self._head, "parameters", mdlpar )

        self.alterParameterNames( dnp )

        self.ncomp -= 1

        return True

    def shuffle( self, param, offset, np, rng ) :
        """
        Shuffle the parameters of the components (if they are equivalent)

        Parameters
        ----------
        param : array-like
            list of all parameters
        offset : int
            index where the dynamic model starts
        np : int
            length of the parameters of the dynamic model
        rng : RNG
            random number generator
        """
        if self.ncomp <= 1 :
            return param

        off = self.model.npbase + offset
        src = numpy.arange( self.ncomp ) * self.deltaNpar + off
        per = rng.permutation( self.ncomp )
        for i in self.index :
            src[-1] = offset + i             # replace last src by index
            param[src[per]] = param[src]
            src += 1

        return param


    #  *************************************************************************
    def __setattr__( self, name, value ) :

        if self.setDynamicAttribute( name, value ) :
            return

        if name == "isDyna" :
            setatt( self, name, value, type=bool )
        elif name == "model" :
            setatt( self, name, value, type=Model )
        elif name == "same" :
            setatt( self, name, value, type=int, islist=True, isnone=True )
        elif name == "index" :
            setatt( self, name, value, type=int, islist=True, isnone=True )
        else :
            super( RepeatingModel, self ).__setattr__( name, value )

    def isDynamic( self ) :
        return self.isDyna

    #  *************************************************************************
    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
             values for the parameters

        """
        result = numpy.zeros( Tools.length( xdata ), dtype=float )
        if self.ncomp == 0 :
            return result

        ke = self.model.npbase
        pars = params[:ke].copy()
        k = 1

#        print( "RM   ", self.ncomp, fmt( params, max=None ) )
        while True :
            result += self.model.result( xdata, pars )
            if k == self.ncomp :
                return result
            k += 1
            ks = ke
            ke += self.deltaNpar
#            print( "RM1  ", k, ks, ke, self.index )
            pars[self.index] = params[ks:ke]

        return result

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
            values for the parameters
        parlist : array_like
            list of indices of active parameter

        """
        nxdata = Tools.length( xdata )
        partial = numpy.zeros( ( nxdata, self.npbase ), dtype=float )
        if self.ncomp == 0 :
            return partial

        ke = self.model.npbase
        pars = params[:ke].copy()
        full = numpy.arange( self.model.npbase, dtype=int )
        k = 1
        while True :
            partial[:,full] += self.model.partial( xdata, pars )
            if k == self.ncomp :
                break
            k += 1
            ks = ke
            ke += self.deltaNpar

            full[self.index] = [k for k in range( ks, ke )]
            pars[self.index] = params[ks:ke]

        if parlist is None :
            return partial

        return partial[:,parlist]

    def baseDerivative( self, xdata, params ):
        """
        Returns the derivative df/dx at the input value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the result
        params : array_like
             values for the parameters

        """
        dfdx = numpy.zeros( Tools.length( xdata ), dtype=float )
        if self.ncomp == 0 :
            return dfdx

        ke = self.model.npbase
        pars = params[:ke].copy()
        k = 1
        while True :
            dfdx += self.model.derivative( xdata, pars )
            if k == self.ncomp :
                return dfdx
            k += 1
            ks = ke
            ke += self.deltaNpar
            pars[self.index] = params[ks:ke]

        return dfdx

    def setLimits( self, lowLimits=None, highLimits=None ) :
        self.model.setLimits( lowLimits=lowLimits, highLimits=highLimits )

    def hasPriors( self ) :
        return self.model.hasPriors()

    def basePrior( self, k ):
        """
        Return the prior for parameter k.

        Parameters
        ----------
        k : int
            the parameter to be selected.
        """
        i,k = self.par2model( k )
        return self.model.getPrior( i )

    def baseName( self ):
        """ Return a string representation of the model.  """
        return ( "Repeat %d times:\n  " % self.ncomp + self.model.baseName( ) )

    def baseParameterName( self, k ):
        """
        Return the name of the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.

        """
        i,k = self.par2model( k )
        return self.model.getParameterName( i ) + "_%d" % k

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.

        """
        self.model.xUnit = self.xUnit
        self.model.yUnit = self.yUnit

        i,k = self.par2model( k )
        return self.model.getParameterUnit( i )

    def par2model( self, k ) :
        """
        Return index in model and repetition nr for param k
        """
        if k >= self.model.npchain :
            k -= self.model.npchain
            i = self.index[ k % self.deltaNpar ]
            k = 1 + k // self.deltaNpar
        else :
            i = k
            k = 0
        return i, k



# run with : python3 -m unittest TestBasicSplinesModel

import unittest
import os
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt
from numpy.testing import assert_array_almost_equal as assertAAE
from numpy.testing import assert_array_equal as assertAE

from BayesicFitting import formatter as fmt
from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """
    def __init__( self, name ):
        super( ).__init__( name )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def test1( self ) :
        print( "==== test 1 ====================" )

        x = numpy.linspace( 0, 10, 101, dtype=float )
        kn1 = numpy.asarray( [0,2,4,6,8,10], dtype=float )

        cc = ['k-', 'b-', 'r-', 'g-', 'c-', 'm-']
        for k in [0,1] :
            sm = BasicSplinesModel( knots=kn1, border=k )

            par = numpy.ones( sm.npars, dtype=float )

            y = sm.result( x, par )
            dy = sm.derivative( x, par )
            pt = sm.partial( x, par )

#            for i in range( 101 ) :
#                print( fmt( x[i] ), fmt( pt[i,:], max=None ) )

            if self.doplot :
                plt.plot( x, y, cc[k] )
                plt.plot( x, dy, cc[k+3] )
                for i in range( sm.npars ) :
                    plt.plot( x, pt[:,i] )

                plt.show()

    def test2( self ) :
        print( "==== test 2 ====================" )

        x = numpy.linspace( 0, 10, 101, dtype=float )
        kn1 = numpy.asarray( [0,1,4,5,6,7,8,10], dtype=float )

        cc = ['k-', 'b-', 'r-', 'g-', 'c-', 'm-']
        for k in [0,1] :
            sm = BasicSplinesModel( knots=kn1, border=k )

            par = numpy.ones( sm.npars, dtype=float )

            y = sm.result( x, par )
            dy = sm.derivative( x, par )
            pt = sm.partial( x, par )

#            for i in range( 101 ) :
#                print( fmt( x[i] ), fmt( pt[i,:], max=None ) )

            if self.doplot :
                plt.plot( x, y, cc[k] )
                plt.plot( x, dy, cc[k+3] )
                for i in range( sm.npars ) :
                    plt.plot( x, pt[:,i] )
                plt.plot( kn1, [-0.02]*8, 'r|' )
                plt.show()

    def test3( self ) :
        print( "==== test 3 BasicSplinesModel ====================" )

        x = numpy.linspace( 0, 100, 1001, dtype=float )

        kn = [0.000, 1.470, 16.049, 28.465, 36.358, 36.374, 36.829,
              59.589, 59.699, 70.193, 75.844, 88.918, 88.919,  100.000]

        sm = BasicSplinesModel( knots=kn )
        par = numpy.ones( sm.npars, dtype=float )

        y = sm.result( x, par )
        dy = sm.derivative( x, par )
        pt = sm.partial( x, par )

        fm = BasicSplinesModel( knots=kn, fixed={0:0.0, -1:0.0} )
        far = numpy.ones( fm.npars, dtype=float )

        print( fm.npbase, fm.npars, fm.npmax, len( far ) )
        print( fm.parlist )
        assertAE( fm.parlist, numpy.arange( 14, dtype=int ) + 1 )
        print( fm.fixed )
        assertAE( list( fm.fixed.keys() ), [0, 15] )


        fy = fm.result( x, far )
        fdy = fm.derivative( x, far )
        fpt = fm.partial( x, far )


        cc = ['k-', 'b-', 'r-', 'g-', 'c-', 'm-']
        if self.doplot :
            for k in kn :
                plt.plot( [k,k], [0,1], 'k:' )

            plt.plot( x, y, cc[0] )
            plt.plot( x, dy, cc[3] )
            for i in range( sm.npars ) :
                plt.plot( x, pt[:,i] )

            plt.plot( x, fy+2, cc[0] )
            plt.plot( x, fdy+2, cc[3] )
            for i in range( fm.npars ) :
                plt.plot( x, fpt[:,i]+2 )

            plt.ylim( -0.3, 3.3 )
            plt.show()

    def test4a( self ) :
        print( "==== test 4a SplinesModel ====================" )

        x = numpy.linspace( 0, 10, 1001, dtype=float )
        kn1 = numpy.linspace( 0, 10, 11 , dtype=float )

        sm1 = SplinesModel( knots=kn1 )
        par1 = numpy.ones( sm1.npars, dtype=float )

        for k in range( 100 ) :
            y1  = sm1.result( x, par1 + k )

    def test4b( self ) :
        print( "==== test 4b BasicSplinesModel ====================" )

        x = numpy.linspace( 0, 10, 1001, dtype=float )
        kn1 = numpy.linspace( 0, 10, 11 , dtype=float )

        sm1 = BasicSplinesModel( knots=kn1 )
        par1 = numpy.ones( sm1.npars, dtype=float )

        for k in range( 100 ) :
            y1  = sm1.result( x, par1 + k )


    def test4c( self ) :
        print( "==== test 4c BSplinesModel ====================" )

        x = numpy.linspace( 0, 10, 1001, dtype=float )
        kn1 = numpy.linspace( 0, 10, 11 , dtype=float )

        sm2 = BSplinesModel( knots=kn1 )
        par2 = numpy.ones( sm2.npars, dtype=float )

        for k in range( 100 ) :
            y1  = sm2.result( x, par2 + k )



    def test4( self ) :
        print( "==== test 4 ====================" )

        x = numpy.linspace( 0, 10, 101, dtype=float )
#        kn1 = numpy.asarray( [0,1,4,5,6,7,8,10], dtype=float )
        kn1 = numpy.linspace( 0, 10, 11 , dtype=float )

        cc = ['k-', 'b-', 'r-', 'g-', 'c-', 'm-']

        sm1 = BasicSplinesModel( knots=kn1 )
        sm2 = BSplinesModel( knots=kn1 )

        print( sm1.npars, sm2.npars )
        self.assertTrue( sm1.npars == sm2.npars )

        par1 = numpy.ones( sm1.npars, dtype=float )
        par2 = numpy.ones( sm2.npars, dtype=float )

        y1  = sm1.result( x, par1 )
        dy1 = sm1.derivative( x, par1 )
        pt1 = sm1.partial( x, par1 )

        y2  = sm2.result( x, par2 )
        dy2 = sm2.derivative( x, par2 )
        pt2 = sm2.partial( x, par2 )

        assertAAE( y1, y2 )
        assertAAE( dy1, dy2 )
        assertAAE( pt1, pt2 )

        if self.doplot :
            k = 0
            plt.plot( x, y1, cc[k] )
            plt.plot( x, y2 + 2, cc[k] )
            for i in range( sm1.npars ) :
                plt.plot( x, pt1[:,i] )
            for i in range( sm2.npars ) :
                plt.plot( x, pt2[:,i] + 2 )
            plt.plot( x, dy1, cc[k+3] )
            plt.plot( x, dy2 + 2, cc[k+3] )

            plt.show()

    def test6c( self ) :
        kn = numpy.asarray( [-10,10], dtype=float )
        for k in range( 5 ) :
            self.bstest5( kn, order=k )

    def test6d( self ) :
        kn = numpy.asarray( [-10, 0, 10], dtype=float )
        self.bstest5( kn )

    def bstest5( self, kn, order=3 ) :
        print( "==== test 6 ====================" )

        x = numpy.linspace( min(kn), max(kn), 101, dtype=float )

        n = 0
        cc = ['k-', 'b-', 'r-', 'g-', 'c-', 'm-']

        sm = BasicSplinesModel( knots=kn, order=order )
        par = numpy.ones( sm.npars, dtype=float )

        ysm = sm.result( x, par )
        pts = sm.partial( x, par )

        print( fmt( numpy.sum( pts, 0 ), max=None ) )

        if self.doplot :

            for k in kn :
                plt.plot( [k,k], [0,1], 'k:' )

            plt.plot( x, ysm, 'k-' )
            for i in range( sm.npars ) :
                plt.plot( x, pts[:,i], cc[3] )

            plt.show()

    def test6a( self ) :
        kn = numpy.linspace( 0, 10, 11, dtype=float )
        self.bstest6( kn )

    def test6a0( self ) :
        kn = numpy.linspace( 0, 10, 11, dtype=float )
        self.bstest6( kn, order=0 )

    def test6a1( self ) :
        kn = numpy.linspace( 0, 10, 11, dtype=float )
        self.bstest6( kn, order=1 )

    def test6a2( self ) :
        kn = numpy.linspace( 0, 10, 11, dtype=float )
        self.bstest6( kn, order=2 )

    def test6a4( self ) :
        kn = numpy.linspace( 0, 10, 11, dtype=float )
        self.bstest6( kn, order=4 )

    def test6b( self ) :
        kn = numpy.asarray( [0,1,2,5,6,8,10], dtype=float )
        self.bstest6( kn )


    def bstest6( self, kn, order=3 ) :
        print( "==== test 6 ====================" )

        x = numpy.linspace( min(kn), max(kn), 101, dtype=float )

        n = 0
        cc = ['k-', 'b-', 'r-', 'g-', 'c-', 'm-']

        sm = BasicSplinesModel( knots=kn, order=order )
        bm = BSplinesModel( knots=kn, order=order )
        par = numpy.ones( bm.npars, dtype=float )

        ysm = sm.result( x, par )
        pts = sm.partial( x, par )

        ybm = bm.result( x, par )
        ptb = bm.partial( x, par )

        print( fmt( numpy.sum( pts, 0 ), max=None ) )
        print( fmt( numpy.sum( ptb, 0 ), max=None ) )

        if self.doplot :

            for k in kn :
                plt.plot( [k,k], [0,1], 'k:' )

            plt.plot( x, ybm, 'b-' )
            plt.plot( x, ysm + 1.1, 'k-' )
            for i in range( bm.npars ) :
#                plt.plot( x, ptb[:,i] - pts[:,i], cc[2] )
                plt.plot( x, ptb[:,i], cc[2] )
                plt.plot( x, pts[:,i]+1.1, cc[3] )

            plt.show()

        assertAAE( ysm, ybm )
        if order > 0 :
            assertAAE( pts, ptb )

    def test7a0( self ) :
        kn = numpy.linspace( 0, 10, 11, dtype=float )
        self.bstest7( kn, order=0 )

    def test7a1( self ) :
        kn = numpy.linspace( 0, 10, 11, dtype=float )
        self.bstest7( kn, order=1 )

    def test7a2( self ) :
        kn = numpy.linspace( 0, 10, 11, dtype=float )
        self.bstest7( kn, order=2 )

    def test7a( self ) :
        kn = numpy.linspace( 0, 10, 11, dtype=float )
        self.bstest7( kn )

    def test7a4( self ) :
        kn = numpy.linspace( 0, 10, 11, dtype=float )
        self.bstest7( kn, order=4 )

    def test7b( self ) :
        kn = numpy.asarray( [0,1,2.1,3,4.4,5,6,8,10], dtype=float )
        self.bstest7( kn )


    def bstest7( self, kn, order=3, plot=False ) :
        print( "==== test 7 ========= order = %d ===========" % ( order ) )

        x = numpy.linspace( 0, 10, 101, dtype=float )

        n = 0
        cc = ['k-', 'b-', 'r-', 'g-', 'c-', 'm-']

        sm = BasicSplinesModel( knots=kn, order=order, border=1 )
        spar = numpy.ones( sm.npars, dtype=float )
        bpar = numpy.linspace( 0.0, 1.0, sm.npars, dtype=float )

        ysm = sm.result( x, spar )
        pts = sm.partial( x, spar )

        ybm = sm.result( x, bpar )
        ptb = sm.partial( x, bpar )

        print( fmt( numpy.sum( pts, 0 ), max=None ) )
        print( fmt( numpy.sum( ptb, 0 ), max=None ) )

        if self.doplot :

            for k in kn :
                plt.plot( [k,k], [0,1], 'k:' )

            plt.plot( x, ybm + 1.1, 'k-' )
            plt.plot( x, ysm, 'k-' )
            for i in range( sm.npars ) :
                plt.plot( x, ptb[:,i] + 1.1, cc[i%6] )
                plt.plot( x, pts[:,i], cc[i%6] )

            plt.show()

        assertAAE( ysm, 1.0 )

    def test8a0( self ) :
        self.bstest8( order=0 )

    def test8a1( self ) :
        self.bstest8( order=1 )

    def test8a2( self ) :
        self.bstest8( order=2 )

    def test8a( self ) :
        self.bstest8( )

    def test8a4( self ) :
        self.bstest8( order=4 )

    def bstest8( self, order=3 ) :
        print( "==== test 8 ========= order %d ===========" % ( order ) )

        x = numpy.linspace( 0, 10, 101, dtype=float )
        cc = ['k-', 'b-', 'r-', 'g-', 'c-', 'm-']

        for k in range( 2, 7 ) :
            knots = numpy.linspace( 0, 10, k, dtype=float )
            print( "Knots  ", fmt( knots, max=None ) )

            bm = BasicSplinesModel( knots=knots, order=order )
            par = numpy.ones( bm.npars, dtype=float )

            ybm = bm.result( x, par )
            ptb = bm.partial( x, par )

            print( fmt( numpy.sum( ptb, 0 ), max=None ) )

            if self.doplot :
                off = 1.1 * ( k - 2 )
                plt.plot( x, ybm + off, 'b-' )
                for i in range( bm.npars ) :
                    plt.plot( x, ptb[:,i] + off, cc[i%6] )

        if self.doplot :
            plt.ylim( -0.1, 6 )
            plt.show()


    def test9( self ) :
        print( "==== test ========= fixed parameters ==========="  )

        fm = BasicSplinesModel( nrknots=8, min=0, max=10, fixed={0:0.0, -1:0.0} )

        print( fm.parlist )
        assertAE( fm.parlist, numpy.arange( 8, dtype=int ) + 1 )
        print( fm.fixed )
        assertAE( list( fm.fixed.keys() ), [0, 9] )

        fm = BasicSplinesModel( nrknots=8, min=0, max=10, fixed={0:0, 1:0, -2:0, -1:0} )

        print( fm.parlist )
        assertAE( fm.parlist, numpy.arange( 6, dtype=int ) + 2 )
        print( fm.fixed )
        assertAE( list( fm.fixed.keys() ), [0, 1, 8, 9] )


    def test10( self ) :
        print( "   Test BasicSplinesModel      " )

        mdl = BasicSplinesModel( knots=[-2.0, 2.1] )

        print( "   Test DynamicModel      " )
        mdl = BasicSplinesModel( knots=[-2.0, 0.0, 2.0] )


    def plot1( self ):
        x = numpy.linspace( 0, 10, 101, dtype=float )
        y = numpy.sin( x )
        kn1 = numpy.linspace( 0, 10, 11, dtype=float )
        sm1 = BasicSplinesModel( knots=kn1 )
        ftr = Fitter( x, sm1 )
        par1 = ftr.fit( y )
        print( fmt( par1, max=None ) )

        plt.plot( x, y, 'k.' )
        plt.plot( x, sm1.result( x, par1 ) )


        plt.show()



    def suite( cls ):
        return unittest.TestCase.suite( Test.__class__ )


if __name__ == '__main__':
    unittest.main( )





## run as : python3 -m unittest TestErrorDistribution

from __future__ import print_function

import numpy as numpy
from numpy.testing import assert_array_almost_equal as assertAAE
import unittest
import os
from astropy import units
import math
import sys
import matplotlib.pyplot as plt
from StdTests import stdModeltest

from BayesicFitting import *
from BayesicFitting import formatter as fmt
#from MultiDimAverageModel import MultiDimAverageModel
#from BoxedErrorDistribution import BoxedErrorDistribution

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2017 - 2021  Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """
    SQRT2 = math.sqrt( 2 )
    LGSQ2 = math.log( SQRT2 )
    LOG2 = math.log( 2 )
    EPSILON = 1e-6

    # Define x independent variable
    x = numpy.asarray( [ -1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0], dtype=float )

    # Define y dependent variable
    data = numpy.asarray( [-11.1, -9.2, -5.1, -5.2, -1.1, 1.2, 1.1, 5.2, 5.1, 8.2, 11.2], dtype=float )

    # Define wgt weigth
    wgt = numpy.asarray( [1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1], dtype=float )

    # Define acc as accuracies
    acc0 = 0.7
    acc1 = 1.0 / numpy.sqrt( wgt )

    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )
        numpy.set_printoptions(formatter={'float': '{: 0.3f}'.format})

    def test1( self ) :
        print( "====test1============================" )
        nn = 10000
        x = numpy.linspace( 0, 1, nn, dtype=float )
        ym = 0.0 * x
        nf = 0.01

        model = PolynomialModel( 0 )
        model.parameters = 0.0
        problem = ClassicProblem( model=model, xdata=x, ydata=ym )
        fitIndex=[0]

        numpy.random.seed( 2345 )
        noise = numpy.random.randn( nn )

        errdis = GaussErrorDistribution( )
        allpars = [0.0, 1.0]
        print( errdis )
        for k in range( 5 ) :
            noise = numpy.random.randn( nn )
            y = ym + nf * noise
            problem.ydata = y
            scale = errdis.getScale( problem, allpars )
            sigma = errdis.toSigma( scale )
            print( fmt( k ), fmt( nf ), fmt( scale ), fmt( sigma ) )
            nf *= 10
        if self.doplot:
            self.ploterrdis( noise, errdis, problem, allpars, num_bins=40 )

        nf = 0.01
        errdis = LaplaceErrorDistribution( )
        print( errdis )
        for k in range( 5 ) :
            noise = numpy.random.laplace( size=nn )
            y = ym + nf * noise
            problem.ydata = y
            scale = errdis.getScale( problem, allpars )
            sigma = errdis.toSigma( scale )
            print( fmt( k ), fmt( nf ), fmt( scale ), fmt( sigma ) )
            nf *= 10
        if self.doplot:
            self.ploterrdis( noise, errdis, problem, allpars )

        nf = 0.01
        errdis = UniformErrorDistribution( )
        print( errdis )
        for k in range( 5 ) :
            noise = 0.5 - numpy.random.rand( nn )
            y = ym + nf * noise
            problem.ydata = y
            scale = errdis.getScale( problem, allpars )
            sigma = errdis.toSigma( scale )
            print( fmt( k ), fmt( nf ), fmt( scale ), fmt( sigma ) )
            nf *= 10
        if self.doplot:
            self.ploterrdis( noise, errdis, problem, allpars )

        nf = 0.01
        errdis = CauchyErrorDistribution ( )
        cp = CauchyPrior()
        print( errdis )
        for k in range( 5 ) :
            noise = numpy.random.rand( nn )
            noise = cp.unit2Domain( noise )
            noise = numpy.maximum( noise, -20 )
            noise = numpy.minimum( noise, +20 )
            y = ym + nf * noise
            problem.ydata = y
            scale = errdis.getScale( problem, allpars )
            sigma = errdis.toSigma( scale )
            print( fmt( k ), fmt( nf ), fmt( scale ), fmt( sigma ) )
            nf *= 10
        if self.doplot :
            self.ploterrdis( noise, errdis, problem, allpars, num_bins=200 )


        nf = 0.01
        errdis = ExponentialErrorDistribution( )
        power = 1.0
        allpars = [0.0, 1.0, 1.0]
        print( errdis, "  power=1" )
        for k in range( 5 ) :
            noise = numpy.random.laplace( size=nn )
            y = ym + nf * noise
            problem.ydata = y
            scale = errdis.getScale( problem, allpars )
            sigma = errdis.toSigma( [scale,power] )
            print( fmt( k ), fmt( nf ), fmt( scale ), fmt( sigma ) )
            nf *= 10
        if self.doplot :
            self.ploterrdis( noise, errdis, problem, allpars, num_bins=200 )

        nf = 0.01
        errdis = ExponentialErrorDistribution( )
        allpars = [0.0, 1.0, 2.0]
        power = 2
        print( errdis, "  power=2"  )
        for k in range( 5 ) :
            noise = numpy.random.randn( nn )
            y = ym + nf * noise
            problem.ydata = y
            scale = errdis.getScale( problem, allpars )
            sigma = errdis.toSigma( [scale,power] )
            print( fmt( k ), fmt( nf ), fmt( scale ), fmt( sigma ) )
            nf *= 10
        if self.doplot :
            self.ploterrdis( noise, errdis, problem, allpars, num_bins=200 )

        nf = 0.01
        power = 10
        errdis = ExponentialErrorDistribution ( )
        allpars = [0.0, 1.0, 10.0]
        print( errdis, "  power=%d" % power  )
        for k in range( 5 ) :
            noise = 2 * numpy.random.rand( nn ) - 1.0
            y = ym + nf * noise
            problem.ydata = y
            scale = errdis.getScale( problem, allpars )
            sigma = errdis.toSigma( [scale,power] )
            print( fmt( k ), fmt( nf ), fmt( scale ), fmt( sigma ) )
            nf *= 10
        if self.doplot :
            self.ploterrdis( noise, errdis, problem, allpars, num_bins=200 )


    def ploterrdis( self, noise, errdis, problem, allpars, num_bins=20 ) :

        plt.hist( noise, num_bins, facecolor='g', alpha=0.5 )
        xx = numpy.linspace( -0.3, 0.3, 601, dtype=float )
        lik = numpy.zeros_like( xx )
        for k,p in enumerate( xx ) :
            allpars[0] = p
            lik[k] = errdis.logLikelihood( problem, allpars )
        maxlik = numpy.max( lik )
        plt.plot( xx, numpy.exp( lik - maxlik ), 'r-' )
        plt.title( errdis.__str__() )
        plt.show()


    def stdtest( self, errdis, problem, param, fitIndex ) :

        chisq = errdis.getChisq( problem, param )
        print( "chisq = %8.3f"%( chisq ) )
        lLdata = errdis.logLdata( problem, param )
        logL0 = numpy.sum( lLdata )

        logL = errdis.logLikelihood( problem, param )
        altL = errdis.logLikelihood_alt( problem, param )
        print( "logL0 ", fmt( logL0 ) )
        print( "logL  ", fmt( logL ) )
        print( "altL  ", fmt( altL ) )
        assertAAE( logL, altL )
        assertAAE( logL, logL0 )

        dL = errdis.partialLogL( problem, param, fitIndex )
        aL = errdis.partialLogL_alt( problem, param, fitIndex )
        nL = errdis.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "partalt = ", aL )
        print( "numpart = ", nL )
        assertAAE( dL, aL, 5 )
        assertAAE( dL, nL, 5 )



    def testGaussErrorDistribution( self ):

        print( "=======   Test Gauss Error Distribution  ==================" )

        poly = PolynomialModel( 1 )
        param = numpy.asarray( [1,10,1], dtype=float )
        ged = GaussErrorDistribution( )
        self.assertTrue( ged.acceptWeight() )

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=self.data,
                  accuracy=self.acc0 )

        #   data = { -11, -9, -5, -5, -1, 1, 1, 5, 5, 8, 11 }
        #   f( x ) = { -10, -8, -6, -4, -2, 0, 2, 4, 6, 8, 10 }
        #   f - d=     1   1  -1   1  -1 -1  1 -1  1  0  -1

        print( "Using variable scale = 1.0 and accuracies" )
        fitIndex = [0,1,-1]
        self.stdtest( ged, problem, param, fitIndex )

        print( "Using variable scale = 0.1 and accuracies" )
        scale = 0.1
        param[2] = scale
        self.stdtest( ged, problem, param, fitIndex )


        print( "Using fixed scale = 1.0 and accuracies" )
        scale = 1.0
        param[2] = scale
        fitIndex = numpy.asarray( [0,1] )
        self.stdtest( ged, problem, param, fitIndex )


        print( "Using fixed scale = 0.5 and accuracies nad weights." )
        scale = 0.5
        param[2] = scale
        problem.weights = self.wgt
        self.stdtest( ged, problem, param, fitIndex )


        scale = 1.0
        for i in range( 11 ) :
            param = numpy.asarray( [i-5,5,1], dtype=float )
            print( param, "  :  ", end="" )
            for k in range( 9 ):
                print( " %8.3f" % ged.logLikelihood( problem, param ), end="" )
                param[1] += 1
            print( "" )


    def testGaussErrorDistribution2( self ):
        print( "\n=====   Test 2 Gauss Error Distribution====================" )
        gm = GaussModel( )
        gm.addModel( PolynomialModel(1) )
        param = numpy.asarray( [5, 1, -0.3, 1, 1, 1], dtype=float )
        data2 = self.data.copy( ) * 0.1
        data2[1:5] = [2.3, 5.4, 4.5, 2.0]
        print( data2 )

        problem = ClassicProblem( model=gm, xdata=self.x, ydata=self.data, accuracy=self.acc1 )
        fitIndex = [0,1,2,3,4,-1]

        ged = GaussErrorDistribution( )
        Tools.printclass( ged )
        np = gm.npchain
        self.stdtest( ged, problem, param, fitIndex )


        scale = 0.1
        param[np] = scale
        self.stdtest( ged, problem, param, fitIndex )


        scale = 1.0
        param[np] = scale
        self.stdtest( ged, problem, param, fitIndex )


        fitIndex = numpy.asarray( [0,2,3,4] )
        self.stdtest( ged, problem, param, fitIndex )


        scale = 0.5
        param[np] = scale
        fitIndex = numpy.asarray( [0,1,2,3,4,-1] )
        problem.weights = self.wgt
        self.stdtest( ged, problem, param, fitIndex )


        scale = 0.5
        param[np] = scale
        fitIndex = numpy.asarray( [0,1,2,4,-1] )
        self.stdtest( ged, problem, param, fitIndex )

        ged.hyperpar[0] = NoiseScale()
        Tools.printclass( ged )
        ged.keepFixed( fixed={0:2.5} )
        Tools.printclass( ged )
        Tools.printclass( ged.hyperpar[0] )

        dL = ged.partialLogL( problem, param, fitIndex )
        nL = ged.numPartialLogL( problem, param, fitIndex )
        print( "partial = ", dL )
        print( "partalt = ", ged.partialLogL_alt( problem, param, fitIndex ) )
        print( "numpart = ", nL )
        assertAAE( dL, nL, 5 )

    def testExponentialErrorDistribution1( self ):

        print( "=======   Test Exponential Error Distribution 1 ==================" )
        ged = ExponentialErrorDistribution( )

        param = numpy.asarray( [1, 10, 0.2, 1.4], dtype=float )
        print( "par   ", param )
        self.stdEDtest( ged, param )


        param = numpy.asarray( [1, 10, 1.2, 3.4], dtype=float )
        print( "par   ", param )
        self.stdEDtest( ged, param )

    def stdEDtest( self, ged, param ) :

        poly = PolynomialModel( 1 )

        print( "---%s with fixed single scale-------" % ged )
        problem = ClassicProblem( model=poly, xdata=self.x, ydata=self.data, accuracy=self.acc0 )
        fitIndex = [0,1]
        self.stdtest( ged, problem, param, fitIndex )

        print( "---%s with fixed single scale and weights-------" %ged )
        problem.weights = self.wgt
        self.stdtest( ged, problem, param, fitIndex )


        print( "---%s with fixed scale per data-------" % ged )
        problem = ClassicProblem( model=poly, xdata=self.x, ydata=self.data, accuracy=self.acc1 )
        self.stdtest( ged, problem, param, fitIndex )

        print( "---%s with fixed scale and weights-------" % ged )
        problem.weights = self.wgt
        self.stdtest( ged, problem, param, fitIndex )


    def testCauchyErrorDistribution( self ):
        print( "\n   Test Cauchy Error Distribution\n" )
        poly = PolynomialModel( 1 )

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=self.data )

        param = numpy.asarray( [1,10,1], dtype=float )

        ced = CauchyErrorDistribution( )

        self.stdEDtest( ced, param ) 


    def testLaplaceErrorDistribution( self ):
        print( "\n   Test Laplace Error Distribution\n" )
        poly = PolynomialModel( 1 )

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=self.data )

        param = numpy.asarray( [1,10,1], dtype=float )
        ced = LaplaceErrorDistribution( )

        self.stdEDtest( ced, param ) 

    def testUniformErrorDistribution( self ):
        print( "\n   Test Uniform Error Distribution\n" )
        poly = PolynomialModel( 1 )

        problem = ClassicProblem( model=poly, xdata=self.x, ydata=self.data )

        self.acc0 *= 5
        self.acc1 *= 5

        ced = UniformErrorDistribution( )

        param = numpy.asarray( [0.3,11,4], dtype=float )
        self.stdEDtest( ced, param ) 



    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( ErrorDistributionTest.__class__ )





import numpy as numpy
from astropy import units
import math
from . import Tools
from .Tools import setAttribute as setatt

from .LinearModel import LinearModel

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2007 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2020 Do Kester

class HarmonicModel( LinearModel ):
    """
    Harmonic oscillator Model.

    f( x:p ) = SUM_j ( p_k * cos( 2*pi*j*x ) + p_k+1 * sin( 2*pi*j*x ) )
    j = 1, N; k = 0, 2N-2.

    The number of parameters is 2 * order.
    The parameters are initialized at 1.0. It is a linear model.

    Author:      Do Kester

    Examples
    --------
    >>> harm = HarmonicModel( 3 )            # period = 1
    >>> print( harm.npbase
    6
    >>> harm = HarmonicModel( 4, 2.7 )        # period = 2.7

    Attributes
    ----------
    order : int
        the order of the harmonic
    period : float
        the length of the period of the fundamental

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    """
    PARNAMES = ["cosamp_", "sinamp_"]

    def __init__( self, order, period=1.0, copy=None, **kwargs ):
        """
        Harmonic oscillator model.

        Number of parameters is 2 * order.

        Parameters
        ----------
        order : int (>0)
            the number of overtones
        period : float
            length of the period of the fundamental. default 1.0
        copy : HarmonicModel
            model to be copied
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """
        super( HarmonicModel, self ).__init__( 2 * order, copy=copy, **kwargs )

        self.order = order
        self.period = period

    def copy( self ):
        """ Copy method.  """
        return HarmonicModel( self.order, period=self.period, copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes: order, period

        """
        if name == 'order' :
            setatt( self, name, value, type=int )
        elif name == 'period' :
            setatt( self, name, value, type=float )
        else :
            super( HarmonicModel, self ).__setattr__( name, value )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            x values at which to calculate the partials
        params : array_like
            parameters of the model. (ignored in LinearModels)
        parlist : array_like
            list of indices of active parameters

        """
        np = self.npmax if parlist is None else len( parlist )
        ni = Tools.length( xdata )
        partial = numpy.zeros( ( ni, np), dtype=float )

        x = 2 * math.pi * xdata / self.period

        parts = { 0 : ( lambda j : numpy.cos( j * x ) ),
                  1 : ( lambda j : numpy.sin( j * x ) ) }

        if parlist is None :
            parlist = range( self.npmax )

        for k,kp in enumerate( parlist ) :
            j = 1 + kp // 2
            partial[:,k] = parts[kp % 2]( j )

        return partial

    def baseDerivative( self, xdata, params ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            value at which to calculate the partials
        params : array_like
            parameters of the model.

        """
        df = numpy.zeros_like( xdata )
        tp = 2 * math.pi / self.period
        x = tp * xdata
        j = 0
        for k in range( 0, self.npmax, 2 ) :
            j += 1
            df += tp * j * ( params[k+1] * numpy.cos( j * x ) - params[k] * numpy.sin( j * x ) )
        return df

    def baseName( self ):
        """ Returns a string representation of the model.  """
        bn = "Harmonic: f( x:p ) = "
        if self.order == 0 :
            bn += "0"
            return bn

        bn += "p_0 * cos( xx ) + p_1 * sin( xx )"
        j = 2
        for k in range( 2, self.order + 1 ) :
            bn += " + p_%d * cos( %d xx ) + p_%d * sin( %d xx )"%(j,k,j+1,k)
            j += 2
        bn += "; xx = 2 * pi * x / period"
        return bn

    def baseParameterName( self, k ):
        """
        Return the name of the indicated parameter.
        Parameters
        ----------
        k : int
            parameter number.

        """
        return "%s%d"%( self.PARNAMES[k%2], k / 2 )

    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.
        Always : YUnit.
        Parameters
        ----------
        k : int
            parameter number.

        """
        return self.yUnit





# run with : python3 -m unittest TestPrior2
# or :       python3 -m unittest TestPrior2.Test.test1

import unittest
import numpy as numpy
import math
import os
import matplotlib.pyplot as plt

from BayesicFitting import Prior, UniformPrior, JeffreysPrior, ExponentialPrior
from BayesicFitting import LaplacePrior, CauchyPrior, GaussPrior
from BayesicFitting import UniformPrior
from BayesicFitting import PolynomialModel, NestedSampler
from BayesicFitting import GaussErrorDistribution

from BayesicFitting import GaussPriorNew
from BayesicFitting import ClassicProblem
from BayesicFitting import Fitter
from BayesicFitting import NestedSampler
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *  2006-2015 Do Kester (JAVA CODE)

class Test( unittest.TestCase ) :
    """
    Test for Priors
    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def addPrior( self, logL, problem, allpars, logLlow ):
        return logL + GaussPrior( center=0.5, scale=0.5/6 ).logResult( allpars[0] )

    def test1( self ) :

        print( "===========Gauss Prior Limitation ===========" )

        NP = 100
        x = numpy.zeros( NP, dtype=float )
        y = 0.4  + 0.2 * numpy.random.randn( NP )

#        print( numpy.mean( numpy.append( y, [5.0] ) ) )

        mdl = PolynomialModel( 0 )

        mdl.setPrior( 0, GaussPrior( center=0.5, scale=0.5/6 ) )

        ed = GaussErrorDistribution( scale=0.2 )
        ns = NestedSampler( x, mdl, y, verbose=2, distribution=ed )
        ns.ensemble = 500

        logE = ns.sample()

        pars = ns.parameters
        stdv = ns.stdevs
        print( pars, stdv, logE )
        maxL = ns.samples[-1].logL
        print( maxL + math.log( stdv ) )

#        lim = math.exp( 4 )
        mdl.setPrior( 0, UniformPrior( limits=[0.0,1.0] ) )

        
        ed = GaussErrorDistribution( scale=0.2 )
        ed.constrain = self.addPrior
        ns = NestedSampler( x, mdl, y, verbose=2, distribution=ed )
        ns.ensemble = 500
        logE = ns.sample()

        pars = ns.parameters
        stdv = ns.stdevs
        print( pars, stdv, logE )
        maxL = ns.samples[-1].logL
        print( maxL + math.log( stdv ) )

    def test2( self ) :

        print( "=============Plot Gauss Prior==============" )
        
        NP = 171
        x = numpy.linspace( -8.5, 8.5, NP, dtype=float )
        u = numpy.linspace( 0, 1, NP, dtype=float )

        gp = GaussPrior( )
        res = gp.result( x )
        print( numpy.sum( res ) / 10 ) 

        if not self.doplot :
            return

        plt.plot( x, res, 'k-' )

        d = []
        for xx in x :
            d += [gp.domain2Unit( xx )]
        d = numpy.array( d )

        plt.plot( x[1:-1], d[1:-1], 'r-' )


        dd = ( d - numpy.roll( d, -2 ) ) / ( x - numpy.roll( x, -2 ) )
        plt.plot( x[3:-1], dd[2:-2], 'g-' )

        gn = GaussPriorNew( )
        res = gp.result( x )
        print( numpy.sum( res ) / 10 ) 

        plt.plot( x, res, 'b-' )

        d = []
        for xx in x :
            d += [gn.domain2Unit( xx )]
        d = numpy.array( d )

        plt.plot( x[1:-1], d[1:-1], 'c-' )


        dd = ( d - numpy.roll( d, -2 ) ) / ( x - numpy.roll( x, -2 ) )
        plt.plot( x[3:-1], dd[2:-2], 'm-' )

        plt.show()

    def testlogL( self ) :

        s = 1 / math.sqrt( 2.0 )
        x = numpy.asarray( [0.0] )      # [2014., 2015., 2016., 2017., 2018., 2019.] )
        y = numpy.asarray( [0.0] )      # [150., 145., 140., 150., 160., 154.] )
        w = numpy.zeros( 1, dtype=float ) + 1 / ( s *s )

        print( w )

        mdl0 = PolynomialModel( 0 )

        prob1 = ClassicProblem( model=mdl0, xdata=x, ydata=y )
        prob2 = ClassicProblem( model=mdl0, xdata=x, ydata=y, weights=w )

        pp = numpy.linspace( -10, 10, 201, dtype=float )

        ed = GaussErrorDistribution( )

        logL1 = numpy.asarray( [ed.logLikelihood( prob1, [p, 1.0] ) for p in pp] )
        logL2 = logL1 + logL1
        logLE = [ed.logLikelihood( prob1, [p, s] ) for p in pp]
        logLW = [ed.logLikelihood( prob2, [p, 1.0] ) for p in pp]

        if not self.doplot :
            return

        plt.plot( pp, numpy.exp( logL1 ), 'k-' )
#        plt.plot( pp, 0.1 + numpy.exp( logL2 ), 'k-' )
#        plt.plot( pp, 0.2 + numpy.exp( logL1 ) ** 2, 'k-' )
        plt.plot( pp, numpy.exp( logLE ), 'r-' )
        plt.plot( pp, numpy.exp( logLW ), 'g-' )

#        plt.plot( pp, ( logL1 ), 'k-' )
#        plt.plot( pp, ( logLE ), 'r-' )
#        plt.plot( pp, ( logLW ), 'g-' )

        plt.xlabel( "residual" )
        plt.ylabel( "likelihood" )

        plt.show()


    def testm0( self ) :
        x = numpy.asarray( [2014., 2015., 2016., 2017., 2018., 2019.] )
        y = numpy.asarray( [155., 145., 140., 150., 160., 154.] )

        x -= 2013

        mdl0 = PolynomialModel( 0 )

        gp = GaussPrior( center=150, scale=20 )
        gn = GaussPriorNew( center=150, scale=20 )
        up = UniformPrior( limits=[130,170] )

        problem = ClassicProblem( model=mdl0, xdata=x, ydata=y )

        w = numpy.zeros( 6, dtype=float ) + 2
        prob2 = ClassicProblem( model=mdl0, xdata=x, ydata=y, weights=w )

        dp = 0.1
        pp = numpy.linspace( 130, 170, 401, dtype=float )

        ed = GaussErrorDistribution( )

        su = 0
        sg = 0.0
        sn = 0.0
        logU = up.logResult( 150 )
        ll = []
        for p in pp :
            logL = ed.logLikelihood( problem, [p, 5.0] )
            logG = gp.logResult( p )
            logN = gn.logResult( p )
            su += math.exp( logL + logU )
            sg += math.exp( logL + logG )
            sn += math.exp( logL + logN )
            ll += [logL]

        plt.plot( pp, numpy.exp( ll ), 'k-' )

        su *= dp
        sg *= dp
        sn *= dp

        print( "1  LogZold  ", fmt( math.log( sg ) ), "  LogZnew  ", fmt( math.log( sn ) ), " LogZuni  ", fmt( math.log( su ) ) )

        mdl0.setPrior( 0, up )
        ftr = Fitter( x, mdl0 )
        par = ftr.fit( y )
        ftr.fixedScale = 5.0        

        print( fmt( par ), fmt( ftr.logZ ), fmt( ftr.getLogZ( limits=[130,170] ) ) )

        ed.scale = 5.0
        engs = ["gibbs", "chord"]

        ns = NestedSampler( problem=prob2, engines=engs, distribution=ed, verbose=1 )
        ev2 = ns.sample()
       
        ns = NestedSampler( problem=problem, engines=engs, distribution=ed, verbose=1 )

        ev = ns.sample()
        print( fmt( ev ), fmt( ev2 ) )

        sl = ns.samples
        pml = numpy.average( [sl[k].parameters for k in range( -100, 0 )] )

        pev = sl.getParameterEvolution( kpar=0 )
        lev = sl.getLogLikelihoodEvolution()

        if self.doplot :
            plt.plot( pev, numpy.exp( lev ), 'r.' )

        print( fmt( ns.parameters ), fmt( pml ), fmt( ns.logZ ), " +- ", fmt( ns.logZprecision ) )

        print( fmt( math.log( su ) / ns.logZ ) )

        indx = numpy.argsort( pev )

        ip = pev[indx]
        il = lev[indx]

        print( fmt( ip ), fmt( numpy.roll( ip, -1 ) ) )

        dip = numpy.roll( ip, -1 )[:-1] - ip[:-1]
        print( fmt( dip ) )

        print( fmt( numpy.sum( dip ) ) )
        lin = numpy.sum( dip * numpy.exp( il[:-1] + logU ) )
        print( fmt( math.log( lin ) ) )
        print( fmt( math.log( numpy.sum( numpy.exp( sl.getLogWeightEvolution() ) ) ) ) )

#        lz = -sys.float_info.max

        if self.doplot :
            plt.show()

    def testm1( self ) :
        x = numpy.asarray( [2014., 2015., 2016., 2017., 2018., 2019.] )
        y = numpy.asarray( [155., 145., 140., 150., 160., 154.] )

        x -= 2013

        mdl = PolynomialModel( 1 )

        gp0 = GaussPrior( center=150, scale=20 )
        gn0 = GaussPriorNew( center=150, scale=20 )
        gp1 = GaussPrior( center=0, scale=20 )
        gn1 = GaussPriorNew( center=0, scale=20 )
        up0 = UniformPrior( limits=[130,170] )
        up1 = UniformPrior( limits=[-20,20] )

        problem = ClassicProblem( model=mdl, xdata=x, ydata=y )

        NPP = 71
        nh = NPP // 2

        pp0 = numpy.linspace( 150 - nh, 150 + nh, NPP, dtype=float )
        pp1 = numpy.linspace( -nh, nh, NPP, dtype=float )

        ed = GaussErrorDistribution( )

        sg = 0.0
        sn = 0.0
        su = 0.0
        logU0 = up0.logResult( 150 )
        logU1 = up1.logResult( 0 )
        for p0 in pp0 :
            logG0 = gp0.logResult( p0 )
            logN0 = gn0.logResult( p0 )
            for p1 in pp1 :
                logL = ed.logLikelihood( problem, [p0, p1, 5.0] )
                logG1 = gp1.logResult( p1 )
                logN1 = gn1.logResult( p1 )
                sg += math.exp( logL + logG0 + logG1 )
                sn += math.exp( logL + logN0 + logN1 )
                su += math.exp( logL + logU0 + logU1 )


        print( "2  LogZold  ", fmt( math.log( sg) ), "  LogZnew  ", fmt( math.log( sn ) ), " LogZuni  ", fmt( math.log( su ) ) )

        mdl.setPrior( 0, up0 )
        mdl.setPrior( 1, up1 )

        ftr = Fitter( x, mdl )
        par = ftr.fit( y )
        ftr.fixedScale = 5.0        

        print( fmt( par ), fmt( ftr.logZ ), fmt( ftr.getLogZ( limits=[[130,-20],[170,20]] ) ) )

        ed.scale = 5.0
        engs = ["gibbs", "chord"]
        ns = NestedSampler( problem=problem, distribution=ed, verbose=1, ensemble=400, engines=engs )
        ns.minimumIterations = 1000

        ev = ns.sample()

        print( fmt( ns.parameters ), fmt( ns.logZ ), " +- ", fmt( ns.logZprecision ) )


    def testm2( self ) :
        x = numpy.asarray( [2014., 2015., 2016., 2017., 2018., 2019.] )
        y = numpy.asarray( [155., 145., 140., 150., 160., 154.] )

        x -= 2013

        mdl = PolynomialModel( 2 )

        gp0 = GaussPrior( center=150, scale=20 )
        gn0 = GaussPriorNew( center=150, scale=20 )
        gp1 = GaussPrior( center=0, scale=20 )
        gn1 = GaussPriorNew( center=0, scale=20 )

#        mdl0.setPrior( 0, GaussPrior( center=150, scale=20 ) )

        problem = ClassicProblem( model=mdl, xdata=x, ydata=y )

        NPP = 71
        nh = NPP // 2

        pp0 = numpy.linspace( 150 - nh, 150 + nh, NPP, dtype=float )
        pp1 = numpy.linspace( -nh, nh, NPP, dtype=float )
        pp2 = numpy.linspace( -nh, nh, NPP, dtype=float )

        ed = GaussErrorDistribution( )

        sg = 0.0
        sn = 0.0
        for p0 in pp0 :
            for p1 in pp1 :
                for p2 in pp2 :
                    logL = ed.logLikelihood( problem, [p0, p1, p2, 5.0] )
                    logG0 = gp0.logResult( p0 )
                    logN0 = gn0.logResult( p0 )
                    logG1 = gp1.logResult( p1 )
                    logN1 = gn1.logResult( p1 )
                    logG2 = gp1.logResult( p2 )
                    logN2 = gn1.logResult( p2 )
                    sg += math.exp( logL + logG0 + logG1 + logG2 )
                    sn += math.exp( logL + logN0 + logN1 + logN2 )

        print( "3  LogZold  ", fmt( math.log( sg) ), "  LogZnew  ", fmt( math.log( sn ) ) )

    def testm3( self ) :
        x = numpy.asarray( [2014., 2015., 2016., 2017., 2018., 2019.] )
        y = numpy.asarray( [155., 145., 140., 150., 160., 154.] )

        x -= 2013

        mdl = PolynomialModel( 3 )

        gp0 = GaussPrior( center=150, scale=20 )
        gn0 = GaussPriorNew( center=150, scale=20 )
        gp1 = GaussPrior( center=0, scale=20 )
        gn1 = GaussPriorNew( center=0, scale=20 )

#        mdl0.setPrior( 0, GaussPrior( center=150, scale=20 ) )

        problem = ClassicProblem( model=mdl, xdata=x, ydata=y )

        NPP = 71
        nh = NPP // 2

        pp0 = numpy.linspace( 150 - nh, 150 + nh, NPP, dtype=float )
        pp1 = numpy.linspace( -nh, nh, NPP, dtype=float )
        pp2 = numpy.linspace( -nh, nh, NPP, dtype=float )
        pp3 = numpy.linspace( -nh, nh, NPP, dtype=float )

        ed = GaussErrorDistribution( )

        sg = 0.0
        sn = 0.0
        for p0 in pp0 :
            logG0 = gp0.logResult( p0 )
            logN0 = gn0.logResult( p0 )
            for p1 in pp1 :
                logG1 = gp1.logResult( p1 )
                logN1 = gn1.logResult( p1 )
                for p2 in pp2 :
                    logG2 = gp1.logResult( p2 )
                    logN2 = gn1.logResult( p2 )
                    for p3 in pp3 :
                        logL = ed.logLikelihood( problem, [p0, p1, p2, p3, 5.0] )
                        logG3 = gp1.logResult( p3 )
                        logN3 = gn1.logResult( p3 )
                        sg += math.exp( logL + logG0 + logG1 + logG2 + logG3 )
                        sn += math.exp( logL + logN0 + logN1 + logN2 + logN3 )

        print( "4  LogZold  ", fmt( math.log( sg) ), "  LogZnew  ", fmt( math.log( sn ) ) )


    @classmethod
    def suite( cls ):
        return ConfiguredTestCase.suite( TestPrior2.__class__ )

if __name__ == '__main__':
    unittest.main( )


